# coding:utf-8
#
# unike/module/model/RESCAL.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 31, 2023
# 
# 该头文件定义了 RESCAL.

"""
RESCAL - 一个张量分解模型。
"""

import torch
import typing
import torch.nn as nn
from .Model import Model
from typing_extensions import override

class RESCAL(Model):

	"""
	``RESCAL`` :cite:`RESCAL` 提出于 2011 年，是很多张量分解模型的基石，模型较复杂。

	评分函数为:

	.. math::
	
		-\mathbf{h}^T \mathbf{M}_r \mathbf{t}

	正三元组的评分函数的值越小越好，如果想获得更详细的信息请访问 :ref:`RESCAL <rescal>`。

	例子::

		from unike.utils import WandbLogger
		from unike.data import KGEDataLoader, BernSampler, TradTestSampler
		from unike.module.model import RESCAL
		from unike.module.loss import MarginLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		wandb_logger = WandbLogger(
			project="unike",
			name="RESCAL-FB15K237",
			config=dict(
				in_path = '../../benchmarks/FB15K237/',
				batch_size = 2048,
				neg_ent = 25,
				test = True,
				test_batch_size = 10,
				num_workers = 16,
				dim = 50,
				margin = 1.0,
				use_gpu = True,
				device = 'cuda:0',
				epochs = 1000,
				lr = 0.1,
				opt_method = 'adagrad',
				valid_interval = 100,
				log_interval = 100,
				save_interval = 100,
				save_path = '../../checkpoint/rescal.pth'
			)
		)
		
		config = wandb_logger.config
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = config.in_path, 
			batch_size = config.batch_size,
			neg_ent = config.neg_ent,
			test = config.test,
			test_batch_size = config.test_batch_size,
			num_workers = config.num_workers,
			train_sampler = BernSampler,
			test_sampler = TradTestSampler
		)
		
		# define the model
		rescal = RESCAL(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = config.dim
		)
		
		# define the loss function
		model = NegativeSampling(
			model = rescal, 
			loss = MarginLoss(margin = config.margin)
		)
		
		# test the model
		tester = Tester(model = rescal, data_loader = dataloader, use_gpu = config.use_gpu, device = config.device)
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(), epochs = config.epochs,
			lr = config.lr, opt_method = config.opt_method, use_gpu = config.use_gpu, device = config.device,
			tester = tester, test = config.test, valid_interval = config.valid_interval,
			log_interval = config.log_interval, save_interval = config.save_interval,
			save_path = config.save_path, wandb_logger = wandb_logger)
		trainer.run()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim: int = 100):

		"""创建 RESCAL 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim: 实体和关系嵌入向量的维度
		:type dim: int
		"""

		super(RESCAL, self).__init__(ent_tol, rel_tol)

		#: 实体和关系嵌入向量的维度
		self.dim: int = dim
		#: 根据实体个数，创建的实体嵌入
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim)
		#: 根据关系个数，创建的关系矩阵
		self.rel_matrices: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim * self.dim)

		nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
		nn.init.xavier_uniform_(self.rel_matrices.weight.data)

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head_emb, tail_emb = self.tri2emb(triples, negs, mode)
		rel_matric = self.rel_matrices(triples[:, 1])
		score = self._calc(head_emb, rel_matric, tail_emb)
		return score

	@override
	def tri2emb(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> tuple[torch.Tensor, torch.Tensor]:

		"""
		返回三元组对应的嵌入向量。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 头实体和尾实体的嵌入向量
		:rtype: tuple[torch.Tensor, torch.Tensor]
		"""
		
		if mode == "single":
			head_emb = self.ent_embeddings(triples[:, 0]).unsqueeze(1)
			tail_emb = self.ent_embeddings(triples[:, 2]).unsqueeze(1)
			
		elif mode == "head-batch" or mode == "head_predict":
			if negs is None:
				head_emb = self.ent_embeddings.weight.data.unsqueeze(0)
			else:
				head_emb = self.ent_embeddings(negs)
				
			tail_emb = self.ent_embeddings(triples[:, 2]).unsqueeze(1)
			
		elif mode == "tail-batch" or mode == "tail_predict": 
			head_emb = self.ent_embeddings(triples[:, 0]).unsqueeze(1)
			
			if negs is None:
				tail_emb = self.ent_embeddings.weight.data.unsqueeze(0)
			else:
				tail_emb = self.ent_embeddings(negs)
		
		return head_emb, tail_emb

	def _calc(
		self,
		h: torch.Tensor,
		r: torch.Tensor,
		t: torch.Tensor) -> torch.Tensor:

		"""计算 RESCAL 的评分函数。
		
		:param h: 头实体的向量。
		:type h: torch.Tensor
		:param r: 关系矩阵。
		:type r: torch.Tensor
		:param t: 尾实体的向量。
		:type t: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""
		
		r = r.view(-1, self.dim, self.dim)
		r = r.unsqueeze(dim=1)
		h = h.unsqueeze(dim=-2)
		hr = torch.matmul(h, r)
		hr = hr.squeeze(dim=-2)
		return -torch.sum(hr * t, -1)

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode) -> torch.Tensor:
		
		"""RESCAL 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		head_emb, tail_emb = self.tri2emb(triples, mode=mode)
		rel_matric = self.rel_matrices(triples[:, 1])
		score = self._calc(head_emb, rel_matric, tail_emb)
		return -score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_head_emb, pos_tail_emb = self.tri2emb(pos_sample)
		pos_rel_transfer = self.rel_matrices(pos_sample[:, 1])
		if mode == "bern":
			neg_head_emb, neg_tail_emb = self.tri2emb(neg_sample)
		else:
			neg_head_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)
		neg_rel_transfer = self.rel_matrices(pos_sample[:, 1])

		pos_regul = (torch.mean(pos_head_emb ** 2) +
					 torch.mean(pos_tail_emb ** 2) +
					 torch.mean(pos_rel_transfer ** 2)) / 3

		neg_regul = (torch.mean(neg_head_emb ** 2) +
					 torch.mean(neg_tail_emb ** 2) +
					 torch.mean(neg_rel_transfer ** 2)) / 3

		regul = (pos_regul + neg_regul) / 2
		
		return regul

def get_rescal_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`RESCAL` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'RESCAL'
			},
			'dim': {
				'values': [50, 100, 200]
			}
		}

	:returns: :py:class:`RESCAL` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'RESCAL'
		},
		'dim': {
			'values': [50, 100, 200]
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/model/Analogy.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 31, 2024
# 
# 该头文件定义了 Analogy.

"""
Analogy 类 - DistMult、HolE 和 ComplEx 的集大成者，效果与 HolE、ComplEx 差不多。
"""

import torch
import typing
import torch.nn as nn
from .Model import Model
from typing_extensions import override

class Analogy(Model):

	"""
	``Analogy`` :cite:`ANALOGY` 提出于 2017 年，:py:class:`unike.module.model.DistMult`、:py:class:`unike.module.model.HolE` 和 :py:class:`unike.module.model.ComplEx` 的集大成者，
	效果与 :py:class:`unike.module.model.HolE`、:py:class:`unike.module.model.ComplEx` 差不多。 

	评分函数为:

	.. math::

	    <\operatorname{Re}(\mathbf{h_c}),\operatorname{Re}(\mathbf{r_c}),\operatorname{Re}(\mathbf{t_c})>
	        +<\operatorname{Re}(\mathbf{h_c}),\operatorname{Im}(\mathbf{r_c}),\operatorname{Im}(\mathbf{t_c})>
	        +<\operatorname{Im}(\mathbf{h_c}),\operatorname{Re}(\mathbf{r_c}),\operatorname{Im}(\mathbf{t_c})>
	        -<\operatorname{Im}(\mathbf{h_c}),\operatorname{Im}(\mathbf{r_c}),\operatorname{Re}(\mathbf{t_c})>
	        +<\mathbf{h_d}, \mathbf{r_d}, \mathbf{t_d}>

	评分函数为 :py:class:`unike.module.model.DistMult` 和 :py:class:`unike.module.model.ComplEx` 两者评分函数的和。:math:`< \mathbf{a}, \mathbf{b}, \mathbf{c} >` 为逐元素多线性点积（element-wise multi-linear dot product），
	正三元组的评分函数的值越大越好，负三元组越小越好，如果想获得更详细的信息请访问 :ref:`ANALOGY <analogy>`。

	例子::

		from unike.config import Trainer, Tester
		from unike.module.model import Analogy
		from unike.module.loss import SoftplusLoss
		from unike.module.strategy import NegativeSampling

		# define the model
		analogy = Analogy(
			ent_tol = train_dataloader.get_ent_tol(),
			rel_tol = train_dataloader.get_rel_tol(),
			dim = 200
		)

		# define the loss function
		model = NegativeSampling(
			model = analogy, 
			loss = SoftplusLoss(),
			batch_size = train_dataloader.get_batch_size(), 
			regul_rate = 1.0
		)

		# test the model
		tester = Tester(model = analogy, data_loader = test_dataloader, use_gpu = True, device = 'cuda:1')

		# train the model
		trainer = Trainer(model = model, data_loader = train_dataloader,
			epochs = 2000, lr = 0.5, opt_method = "adagrad", use_gpu = True, device = 'cuda:1',
			tester = tester, test = True, valid_interval = 100,
			log_interval = 100, save_interval = 100,
			save_path = '../../checkpoint/analogy.pth', delta = 0.01)
		trainer.run()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim: int = 100):

		"""创建 Analogy 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim: 实体嵌入向量和关系嵌入向量的维度
		:type dim: int
		"""

		super(Analogy, self).__init__(ent_tol, rel_tol)

		#: 实体嵌入向量和关系嵌入向量的维度
		self.dim: int = dim
		#: 根据实体个数，创建的实体嵌入
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim * 4)
		#: 根据关系个数，创建的关系嵌入
		self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim * 4)
		
		nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
		nn.init.xavier_uniform_(self.rel_embeddings.weight.data)

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
		score = self._calc(head_emb, relation_emb, tail_emb)
		return score

	def _calc(
		self,
		head_emb: torch.Tensor,
		relation_emb: torch.Tensor,
		tail_emb: torch.Tensor) -> torch.Tensor:

		"""计算 Analogy 的评分函数。
		
		:param head_emb: 头实体的向量。
		:type head_emb: torch.Tensor
		:param relation_emb: 关系的向量。
		:type relation_emb: torch.Tensor
		:param tail_emb: 尾实体的向量。
		:type tail_emb: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head, h = torch.chunk(head_emb, 2, dim=-1)
		h_re, h_im = torch.chunk(head, 2, dim=-1)
		relation, r = torch.chunk(relation_emb, 2, dim=-1)
		r_re, r_im = torch.chunk(relation, 2, dim=-1)
		tail, t = torch.chunk(tail_emb, 2, dim=-1)
		t_re, t_im = torch.chunk(tail, 2, dim=-1)

		return (torch.sum(r_re * h_re * t_re +
						  r_re * h_im * t_im +
						  r_im * h_re * t_im -
						  r_im * h_im * t_re, -1)
					+ torch.sum(h * t * r, -1))

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode) -> torch.Tensor:
		
		"""Analogy 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
		score = self._calc(head_emb, relation_emb, tail_emb)
		return score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_head_emb, pos_relation_emb, pos_tail_emb = self.tri2emb(pos_sample)
		if mode == "bern":
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(neg_sample)
		else:
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)

		pos_regul = (torch.mean(pos_head_emb ** 2) + 
					 torch.mean(pos_relation_emb ** 2) + 
					 torch.mean(pos_tail_emb ** 2)) / 3

		neg_regul = (torch.mean(neg_head_emb ** 2) + 
					 torch.mean(neg_relation_emb ** 2) + 
					 torch.mean(neg_tail_emb ** 2)) / 3

		regul = (pos_regul + neg_regul) / 2
		
		return regul

def get_analogy_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`Analogy` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'Analogy'
			},
			'dim': {
				'values': [50, 100, 200]
			}
		}

	:returns: :py:class:`Analogy` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'Analogy'
		},
		'dim': {
			'values': [50, 100, 200]
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/utils/EarlyStopping.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 5, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 6, 2024
#
# 该脚本定义了 EarlyStopping 类.

"""
EarlyStopping - 使用早停止避免过拟合。
"""

import os
import numpy as np
from ..module.model import Model
from loguru import logger

class EarlyStopping:

    """
    如果验证得分（越大越好）在给定的耐心后没有改善，则提前停止训练。
    """

    def __init__(
        self,
        save_path: str,
        patience: int = 2,
        verbose: bool = True,
        delta: float = 0):

        """创建 EarlyStopping 对象。
        
        :param save_path: 模型保存目录
        :type save_path: str
        :param patience: 上次验证得分改善后等待多长时间。默认值：2
        :type patience: int
        :param verbose: 如果为 True，则为每个验证得分改进打印一条消息。默认值：True
        :type verbose: bool
        :param delta: 监测数量的最小变化才符合改进条件。默认值：0
        :type delta: float
        """
        
        #: 模型保存目录
        self.save_path: str = os.path.join(save_path, 'best_network.pth')
        #: 上次验证得分改善后等待多长时间。默认值：2
        self.patience: int = patience
        #: 如果为 True，则为每个验证得分改进打印一条消息。默认值：True
        self.verbose: bool = verbose
        #: 监测数量的最小变化才符合改进条件。默认值：0
        self.delta: float = delta
        
        #: 计数变量
        self.counter: int = 0
        #: 保存最好的得分
        self.best_score: float = -np.Inf
        #: 早停开关
        self.early_stop: bool = False

    def __call__(
        self,
        score: float,
        model: Model):

        """
        进行早停记录。
        """
       
        if score <= self.best_score + self.delta:
            self.counter += 1
            logger.info(f'EarlyStopping counter: {self.counter} / {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.save_checkpoint(score, model)
            self.counter = 0

    def save_checkpoint(
        self,
        score: float,
        model: Model):

        """
        当验证得分改善时保存模型。
        """
        
        if self.verbose:
            logger.info(f'Validation score improved ({self.best_score:.6f} --> {score:.6f}).  Saving model ...')
        model.save_checkpoint(self.save_path)
        self.best_score = score


# coding:utf-8
#
# unike/config/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 3, 2023
# 
# 该头文件定义了 config 接口.

"""循环部分，包含训练循环和验证循环。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .Trainer import Trainer, get_trainer_hpo_config
from .Tester import Tester, get_tester_hpo_config, link_predict, head_predict, tail_predict, calc_ranks

from .HPOTrainer import set_hpo_config, set_hpo_hits,start_hpo_train, hpo_train

__all__ = [
	'Trainer',
	'get_trainer_hpo_config',
	'Tester',
	'get_tester_hpo_config',
	'link_predict',
	'head_predict',
	'tail_predict',
	'calc_ranks',
	'set_hpo_config',
	'set_hpo_hits',
	'start_hpo_train',
	'hpo_train',
]



# coding:utf-8
#
# unike/module/model/RGCN.py
# 
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 22, 2024
# 
# 该头文件定义了 R-GCN.

"""
R-GCN - 第一个图神经网络模型。
"""

import dgl
import torch
import typing
import torch.nn as nn
import torch.nn.functional as F
from .Model import Model
from dgl.nn.pytorch import RelGraphConv
from typing_extensions import override

class RGCN(Model):
    
    """
	``R-GCN`` :cite:`R-GCN` 提出于 2017 年，是第一个图神经网络模型。
	
	正三元组的评分函数的值越大越好，如果想获得更详细的信息请访问 :ref:`R-GCN <rgcn>`。

    例子::

        from unike.data import GraphDataLoader
        from unike.module.model import RGCN
        from unike.module.loss import RGCNLoss
        from unike.module.strategy import RGCNSampling
        from unike.config import Trainer, GraphTester
        
        dataloader = GraphDataLoader(
        	in_path = "../../benchmarks/FB15K237/",
        	batch_size = 60000,
        	neg_ent = 10,
        	test = True,
        	test_batch_size = 100,
        	num_workers = 16
        )
        
        # define the model
        rgcn = RGCN(
        	ent_tol = dataloader.get_ent_tol(),
        	rel_tol = dataloader.get_rel_tol(),
        	dim = 500,
        	num_layers = 2
        )
        
        # define the loss function
        model = RGCNSampling(
        	model = rgcn,
        	loss = RGCNLoss(model = rgcn, regularization = 1e-5)
        )
        
        # test the model
        tester = GraphTester(model = rgcn, data_loader = dataloader, use_gpu = True, device = 'cuda:0')
        
        # train the model
        trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
        	epochs = 10000, lr = 0.0001, use_gpu = True, device = 'cuda:0',
        	tester = tester, test = True, valid_interval = 500, log_interval = 500,
        	save_interval = 500, save_path = '../../checkpoint/rgcn.pth'
        )
        trainer.run()
	"""

    def __init__(
        self,
        ent_tol: int,
        rel_tol: int,
        dim: int,
        num_layers: int):

        """创建 RGCN 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim: 实体和关系嵌入向量的维度
		:type dim: int
		:param num_layers: 图神经网络的层数
		:type num_layers: int
		"""

        super(RGCN, self).__init__(ent_tol, rel_tol)

        #: 实体和关系嵌入向量的维度
        self.dim: int = dim
        #: 图神经网络的层数
        self.num_layers: int = num_layers

        #: 根据实体个数，创建的实体嵌入
        self.ent_emb: torch.nn.Embedding = None
        #: 根据关系个数，创建的关系嵌入
        self.rel_emb: torch.nn.parameter.Parameter = None
        #: R-GCN 的图神经网络层
        self.RGCN: torch.nn.ModuleList = None
        #: 图神经网络层的输出
        self.Loss_emb: torch.nn.Embedding = None

        self.build_model()

    def build_model(self):

        """构建模型"""

        self.ent_emb = nn.Embedding(self.ent_tol, self.dim)

        self.rel_emb = nn.Parameter(torch.Tensor(self.rel_tol, self.dim))

        nn.init.xavier_uniform_(self.rel_emb, gain=nn.init.calculate_gain('relu'))

        self.RGCN = nn.ModuleList()
        for idx in range(self.num_layers):
            RGCN_idx = self.build_hidden_layer(idx)
            self.RGCN.append(RGCN_idx)

    def build_hidden_layer(
        self,
        idx: int) -> dgl.nn.pytorch.conv.RelGraphConv:

        """返回第 idx 的图神经网络层。
        
        :param idx: 数据。
        :type idx: int
        :returns: 图神经网络层
        :rtype: dgl.nn.pytorch.conv.RelGraphConv
        """

        act = F.relu if idx < self.num_layers - 1 else None
        return RelGraphConv(self.dim, self.dim, self.rel_tol, "bdd",
                    num_bases=100, activation=act, self_loop=True, dropout=0.2)

    @override  
    def forward(
        self,
        graph: dgl.DGLGraph,
        ent: torch.Tensor,
        rel: torch.Tensor,
        norm: torch.Tensor,
        triples: torch.Tensor,
        mode: str = 'single') -> torch.Tensor:

        """
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
        
        :param graph: 子图
        :type graph: dgl.DGLGraph
        :param ent: 子图的实体
        :type ent: torch.Tensor
        :param rel: 子图的关系
        :type rel: torch.Tensor
        :param norm: 关系的归一化系数
        :type norm: torch.Tensor
        :param triples: 三元组
        :type triples: torch.Tensor
        :param mode: 模式
        :type mode: str
        :returns: 三元组的得分
        :rtype: torch.Tensor
		"""
        
        embedding = self.ent_emb(ent.squeeze())
        for layer in self.RGCN:
            embedding = layer(graph, embedding, rel, norm)
        self.Loss_emb = embedding
        head_emb, rela_emb, tail_emb = self.tri2emb(embedding, triples, mode)
        score = self.distmult_score_func(head_emb, rela_emb, tail_emb, mode)

        return score

    def tri2emb(
        self,
        embedding: torch.Tensor,
        triples: torch.Tensor,
        mode: str = "single") -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        """
		获得三元组对应头实体、关系和尾实体的嵌入向量。
        
        :param embedding: 经过图神经网络更新的实体嵌入向量
        :type embedding: torch.Tensor
        :param triples: 训练的三元组
        :type triples: torch.Tensor
        :param mode: 模式
        :type mode: str
        :returns: 头实体、关系和尾实体的嵌入向量
        :rtype: torch.Tensor
		"""

        rela_emb = self.rel_emb[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
        head_emb = embedding[triples[:, 0]].unsqueeze(1)  # [bs, 1, dim] 
        tail_emb = embedding[triples[:, 2]].unsqueeze(1)  # [bs, 1, dim]

        if mode == "head-batch" or mode == "head_predict":
            head_emb = embedding.unsqueeze(0)  # [1, num_ent, dim]

        elif mode == "tail-batch" or mode == "tail_predict":
            tail_emb = embedding.unsqueeze(0)  # [1, num_ent, dim]

        return head_emb, rela_emb, tail_emb

    def distmult_score_func(
        self,
        head_emb: torch.Tensor,
        relation_emb: torch.Tensor,
        tail_emb: torch.Tensor,
        mode: str) -> torch.Tensor:

        """
		计算 DistMult 的评分函数。
        
        :param head_emb: 头实体嵌入向量
        :type head_emb: torch.Tensor
        :param relation_emb: 关系嵌入向量
        :type relation_emb: torch.Tensor
        :param tail_emb: 尾实体嵌入向量
        :type tail_emb: torch.Tensor
        :returns: 三元组的得分
        :rtype: torch.Tensor
		"""

        if mode == 'head-batch':
            score = head_emb * (relation_emb * tail_emb)
        else:
            score = (head_emb * relation_emb) * tail_emb

        score = score.sum(dim = -1)
        return score

    @override
    def predict(
        self,
        data: dict[str, torch.Tensor],
        mode: str) -> torch.Tensor:

        """R-GCN 的推理方法。
        
        :param data: 数据。
        :type data: dict[str, torch.Tensor]
        :param mode: 模式
        :type mode: str
        :returns: 三元组的得分
        :rtype: torch.Tensor
		"""

        triples    = data['positive_sample']
        graph      = data['graph']
        ent        = data['entity']
        rel        = data['rela']
        norm       = data['norm']

        embedding = self.ent_emb(ent.squeeze())
        for layer in self.RGCN:
            embedding = layer(graph, embedding, rel, norm)
        self.Loss_emb = embedding
        head_emb, rela_emb, tail_emb = self.tri2emb(embedding, triples, mode)
        score = self.distmult_score_func(head_emb, rela_emb, tail_emb, mode)

        return score

def get_rgcn_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`RGCN` 的默认超参数优化配置。
	
	默认配置为::
	
	    parameters_dict = {
	    	'model': {
	    		'value': 'RGCN'
	    	},
	    	'dim': {
	    		'values': [200, 300, 400]
	    	},
	    	'num_layers': {
	    		'value': 2
	    	}
	    }

	:returns: :py:class:`RGCN` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'RGCN'
		},
		'dim': {
			'values': [200, 300, 400]
		},
		'num_layers': {
			'value': 2
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/data/TradTestSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 29, 2024
#
# 平移模型和语义匹配模型的测试数据采样器.

"""
TradTestSampler - 平移模型和语义匹配模型的测试数据采样器。
"""

import torch
from .TradSampler import TradSampler
from .TestSampler import TestSampler
from typing_extensions import override

class TradTestSampler(TestSampler):

    """平移模型和语义匹配模型的测试数据采样器。
    """

    def __init__(
        self,
        sampler: TradSampler,
        valid_file: str = "valid2id.txt",
        test_file: str = "test2id.txt",
        type_constrain: bool = True):

        """创建 TradTestSampler 对象。

        :param sampler: 训练数据采样器。
        :type sampler: TradSampler
        :param valid_file: valid2id.txt
        :type valid_file: str
        :param test_file: test2id.txt
        :type test_file: str
        :param type_constrain: 是否报告 type_constrain.txt 限制的测试结果
        :type type_constrain: bool
        """

        super().__init__(
            sampler=sampler,
            valid_file=valid_file,
            test_file=test_file,
            type_constrain=type_constrain
        )

        self.get_hr2t_rt2h_from_all()

    @override
    def sampling(
        self,
        data: list[tuple[int, int, int]]) -> dict[str, torch.Tensor]:

        """采样函数。
        
        :param data: 测试的正确三元组
        :type data: list[tuple[int, int, int]]
        :returns: 测试数据
        :rtype: dict[str, torch.Tensor]
        """
        
        batch_data = {}
        head_label = torch.zeros(len(data), self.ent_tol)
        tail_label = torch.zeros(len(data), self.ent_tol)
        for idx, triple in enumerate(data):
            head, rel, tail = triple
            head_label[idx][self.rt2h_all[(rel, tail)]] = 1.0
            tail_label[idx][self.hr2t_all[(head, rel)]] = 1.0

        if self.type_constrain:
            head_label_type = torch.ones(len(data), self.ent_tol)
            tail_laebl_type = torch.ones(len(data), self.ent_tol)
            for idx, triple in enumerate(data):
                head, rel, tail = triple
                head_label_type[idx][self.rel_heads[rel]] = 0.0
                tail_laebl_type[idx][self.rel_tails[rel]] = 0.0
                head_label_type[idx][self.rt2h_all[(rel, tail)]] = 1.0
                tail_laebl_type[idx][self.hr2t_all[(head, rel)]] = 1.0
            batch_data["head_label_type"] = head_label_type
            batch_data["tail_label_type"] = tail_laebl_type
        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["head_label"] = head_label
        batch_data["tail_label"] = tail_label
        return batch_data


# coding:utf-8
#
# unike/data/KGReader.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 17, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
#
# 从文件中读取知识图谱.

"""
KGReader - 从文件中读取知识图谱。
"""

import os
import numpy as np
import collections

class KGReader:

    """
    从文件中读取知识图谱。
    """
    
    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt"):

        """创建 KGReader 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        """

        #: 数据集目录
        self.in_path: str = in_path
        #: entity2id.txt
        self.ent_file: str = ent_file
        #: relation2id.txt
        self.rel_file: str = rel_file
        #: train2id.txt
        self.train_file: str = train_file

        #: 实体的个数
        self.ent_tol: int = 0
        #: 关系的个数
        self.rel_tol: int = 0
        #: 训练集三元组的个数
        self.train_tol: int = 0

        #: 实体->ID
        self.ent2id: dict = {}
        #: 关系->ID
        self.rel2id: dict = {}
        #: ID->实体
        self.id2ent: dict = {}
        #: ID->关系
        self.id2rel: dict = {}

        #: 训练集三元组
        self.train_triples: list[tuple[int, int, int]] = []

        #: 训练集中所有 h-r 对对应的 t 集合
        self.hr2t_train: collections.defaultdict[set] = collections.defaultdict(set)
        #: 训练集中所有 r-t 对对应的 h 集合
        self.rt2h_train: collections.defaultdict[set] = collections.defaultdict(set)

        self.get_id()
        self.get_train_triples_id()
    
    def get_id(self):

        """读取 :py:attr:`ent_file` 文件和 :py:attr:`rel_file` 文件。"""
        
        with open(os.path.join(self.in_path, self.ent_file)) as f:
            self.ent_tol = (int)(f.readline())
            for line in f:
                entity, eid = line.strip().split("\t")
                self.ent2id[entity] = int(eid)
                self.id2ent[int(eid)] = entity
        
        with open(os.path.join(self.in_path, self.rel_file)) as f:
            self.rel_tol = (int)(f.readline())
            for line in f:
                relation, rid = line.strip().split("\t")
                self.rel2id[relation] = int(rid)
                self.id2rel[int(rid)] = relation
    
    def get_train_triples_id(self):

        """读取 :py:attr:`train_file` 文件。"""
        
        with open(os.path.join(self.in_path, self.train_file)) as f:
            self.train_tol = (int)(f.readline())
            for line in f:
                h, t, r = line.strip().split()
                self.train_triples.append((int(h), int(r), int(t)))
        
    def get_hr2t_rt2h_from_train(self):

        """获得 :py:attr:`hr2t_train` 和 :py:attr:`rt2h_train` 。"""
        
        for h, r, t in self.train_triples:
            self.hr2t_train[(h, r)].add(t)
            self.rt2h_train[(r, t)].add(h)
        for h, r in self.hr2t_train:
            self.hr2t_train[(h, r)] = np.array(list(self.hr2t_train[(h, r)]))
        for r, t in self.rt2h_train:
            self.rt2h_train[(r, t)] = np.array(list(self.rt2h_train[(r, t)]))
            
    def get_hr_train(self):

        """用于 ``CompGCN`` :cite:`CompGCN` 训练，因为 ``CompGCN`` :cite:`CompGCN` 的组合运算仅需要头实体和关系。
        
        如果想获得更详细的信息请访问 :ref:`CompGCN <compgcn>`。
        """
        
        self.t_triples = self.train_triples
        self.train_triples = [(hr, list(t)) for (hr,t) in self.hr2t_train.items()]
        
    def get_train(self) -> list[tuple[int, int, int]]:

        """
        返回训练集三元组。

        :returns: :py:attr:`train_triples`
        :rtype: list[tuple[int, int, int]]
        """

        return self.train_triples


# coding:utf-8
#
# unike/module/loss/RGCN_Loss.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 17, 2023
#
# 该脚本定义了 RGCNLoss 类.

"""
RGCNLoss - 损失函数类，R-GCN 原论文中应用这种损失函数完成模型学习。
"""

import torch
from typing import Any
from ..model import RGCN
import torch.nn.functional as F
from .Loss import Loss

class RGCNLoss(Loss):
    
    """
	``R-GCN`` :cite:`R-GCN` 原论文中应用这种损失函数完成模型训练。
	
	.. Note:: :py:meth:`forward` 中的正样本评分函数的得分应大于负样本评分函数的得分。

	例子::

        from unike.module.loss import RGCNLoss
        from unike.module.strategy import RGCNSampling
                
        # define the loss function
        model = RGCNSampling(
        	model = rgcn,
        	loss = RGCNLoss(model = rgcn, regularization = 1e-5)
        )
	"""

    def __init__(
        self,
        model: RGCN,
        regularization: float):

        """创建 RGCNLoss 对象。
        
        :param model: 模型
        :type model: RGCN
        :param regularization: 正则率
        :type regularization: float
		"""
        
        super(RGCNLoss, self).__init__()

        #: 模型
        self.model: RGCN = model
        #: 正则率
        self.regularization: float = regularization
    
    def reg_loss(self) -> torch.Tensor:

        """获得正则部分的损失。
        
        :returns: 损失值
        :rtype: torch.Tensor
		"""

        return torch.mean(self.model.Loss_emb.pow(2)) + torch.mean(self.model.rel_emb.pow(2))

    def forward(
        self,
        score: torch.Tensor,
        labels: torch.Tensor) -> torch.Tensor:

        """计算 RGCNLoss 损失函数。定义每次调用时执行的计算。:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
        
        :param score: 模型的得分。
        :type score: torch.Tensor
        :param labels: 标签
        :type labels: torch.Tensor
        :returns: 损失值
        :rtype: torch.Tensor
		"""

        loss = F.binary_cross_entropy_with_logits(score, labels)
        regu = self.regularization * self.reg_loss()
        loss += regu
        return loss

def get_rgcn_loss_hpo_config() -> dict[str, dict[str, Any]]:

	"""返回 :py:class:`RGCNLoss` 的默认超参数优化配置。
	
	默认配置为::
	
	    parameters_dict = {
	    	'loss': {
	    		'value': 'RGCNLoss'
	    	},
	    	'regularization': {
	    		'value': 1e-5
	    	}
	    }
	
	:returns: :py:class:`RGCNLoss` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'loss': {
			'value': 'RGCNLoss'
		},
		'regularization': {
			'value': 1e-5
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/strategy/RGCNSampling.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 18, 2023
#
# 该脚本定义了 R-GCN 模型的训练策略.

"""
NegativeSampling - 训练策略类，包含损失函数。
"""

import dgl
import torch
import typing
from ..loss import Loss
from ..model import Model
from .Strategy import Strategy

class RGCNSampling(Strategy):

	"""
	将模型和损失函数封装到一起，方便模型训练，用于 ``R-GCN`` :cite:`R-GCN`。

    例子::

        from unike.data import GraphDataLoader
        from unike.module.model import RGCN
        from unike.module.loss import RGCNLoss
        from unike.module.strategy import RGCNSampling
        from unike.config import Trainer, GraphTester
        
        dataloader = GraphDataLoader(
        	in_path = "../../benchmarks/FB15K237/",
        	batch_size = 60000,
        	neg_ent = 10,
        	test = True,
        	test_batch_size = 100,
        	num_workers = 16
        )
        
        # define the model
        rgcn = RGCN(
        	ent_tol = dataloader.train_sampler.ent_tol,
        	rel_tol = dataloader.train_sampler.rel_tol,
        	dim = 500,
        	num_layers = 2
        )
        
        # define the loss function
        model = RGCNSampling(
        	model = rgcn,
        	loss = RGCNLoss(model = rgcn, regularization = 1e-5)
        )
        
        # test the model
        tester = GraphTester(model = rgcn, data_loader = dataloader, use_gpu = True, device = 'cuda:0')
        
        # train the model
        trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
        	epochs = 10000, lr = 0.0001, use_gpu = True, device = 'cuda:0',
        	tester = tester, test = True, valid_interval = 500, log_interval = 500,
        	save_interval = 500, save_path = '../../checkpoint/rgcn.pth'
        )
        trainer.run()
	"""

	def __init__(
		self,
		model: Model = None,
		loss: Loss = None):
		
		"""创建 RGCNSampling 对象。
		
		:param model: R-GCN 模型
		:type model: :py:class:`unike.module.model.RGCN`
		:param loss: 损失函数。
		:type loss: :py:class:`unike.module.loss.Loss`
		"""

		super(RGCNSampling, self).__init__()

		#: R-GCN 模型，即 :py:class:`unike.module.model.RGCN`
		self.model: Model = model
		#: 损失函数，即 :py:class:`unike.module.loss.Loss`
		self.loss: Loss = loss

	def forward(
		self,
		data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]) -> torch.Tensor:
		
		"""计算最后的损失值。定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param data: 数据
		:type data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
		:returns: 损失值
		:rtype: torch.Tensor
		"""
		
		graph    = data["graph"]
		entity   = data['entity']
		relation = data['relation']
		norm     = data['norm']
		triples  = data["triples"]
		label    = data["label"]
		score = self.model(graph, entity, relation, norm, triples)
		loss  = self.loss(score,  label)
		return loss

def get_rgcn_sampling_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`RGCNSampling` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'strategy': {
				'value': 'RGCNSampling'
			}
		}
	
	:returns: :py:class:`RGCNSampling` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'strategy': {
			'value': 'RGCNSampling'
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/data/CompGCNSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 19, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 19, 2023
#
# 该脚本定义了 CompGCNSampler 类.

"""
CompGCNSampler - CompGCN 的数据采样器。
"""

import dgl
import torch
import typing
import numpy as np
from .RGCNSampler import RGCNSampler
from typing_extensions import override

class CompGCNSampler(RGCNSampler):
    
    """``CompGCN`` :cite:`CompGCN` 的训练数据采样器。

    例子::

        from unike.data import RGCNSampler, CompGCNSampler
        from torch.utils.data import DataLoader

        #: 训练数据采样器
        train_sampler: typing.Union[typing.Type[RGCNSampler], typing.Type[CompGCNSampler]] = train_sampler(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file,
            batch_size=batch_size,
            neg_ent=neg_ent
        )

        #: 训练集三元组
        data_train: list[tuple[int, int, int]] = train_sampler.get_train()

        train_dataloader = DataLoader(
            data_train,
            shuffle=True,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=True,
            drop_last=True,
            collate_fn=train_sampler.sampling,
        )
    """

    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt",
        batch_size: int | None = None,
        neg_ent: int = 1):

        """创建 CompGCNSampler 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        :param batch_size: batch size
        :type batch_size: int | None
        :param neg_ent: 对于 CompGCN 不起作用。
        :type neg_ent: int
        """

        super().__init__(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file,
            batch_size=batch_size,
            neg_ent=neg_ent
        )
        
        super().get_hr_train()
        
        self.graph, self.relation, self.norm = \
            self.build_graph(self.ent_tol, np.array(self.t_triples).transpose(), -0.5)

    @override
    def sampling(
        self,
        pos_hr_t: list[tuple[tuple[int, int], list[int]]]) -> dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]:
        
        """``CompGCN`` :cite:`CompGCN` 的采样函数。

        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[tuple[int, int], list[int]]]
        :returns: ``CompGCN`` :cite:`CompGCN` 的训练数据
        :rtype: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
        """

        batch_data = {}
        
        self.label = torch.zeros(self.batch_size, self.ent_tol)
        self.triples  = torch.LongTensor([hr for hr , _ in pos_hr_t])
        for id, hr_sample in enumerate([t for _ ,t in pos_hr_t]):
            self.label[id][hr_sample] = 1

        batch_data['sample']   = self.triples
        batch_data['label']    = self.label
        batch_data['graph']    = self.graph
        batch_data['relation'] = self.relation
        batch_data['norm']     = self.norm

        return batch_data

    @override
    def node_norm_to_edge_norm(
        self,
        graph: dgl.DGLGraph,
        node_norm: torch.Tensor) -> torch.Tensor:

        """根据源节点和目标节点的度计算每条边的归一化系数。

        :param graph: 子图的节点数
        :type graph: dgl.DGLGraph
        :param node_norm: 节点的归一化系数
        :type node_norm: torch.Tensor
        :returns: 边的归一化系数
        :rtype: torch.Tensor
        """
        
        graph.ndata['norm'] = node_norm
        graph.apply_edges(lambda edges: {'norm': edges.dst['norm'] * edges.src['norm']})
        norm = graph.edata.pop('norm').squeeze()
        return norm


# coding:utf-8
#
# unike/data/TestSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 29, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
#
# 测试数据采样器基类.

"""
TestSampler - 测试数据采样器基类。
"""

import os
import torch
import typing
from .TradSampler import TradSampler
from .RGCNSampler import RGCNSampler
from .CompGCNSampler import CompGCNSampler
from collections import defaultdict as ddict
from ..utils import construct_type_constrain

class TestSampler(object):

    """测试数据采样器基类。
    """

    def __init__(
        self,
        sampler: typing.Union[TradSampler, RGCNSampler, CompGCNSampler],
        valid_file: str = "valid2id.txt",
        test_file: str = "test2id.txt",
        type_constrain: bool = True):

        """创建 TestSampler 对象。

        :param sampler: 训练数据采样器。
        :type sampler: typing.Union[TradSampler, RGCNSampler, CompGCNSampler]
        :param valid_file: valid2id.txt
        :type valid_file: str
        :param test_file: test2id.txt
        :type test_file: str
        :param type_constrain: 是否报告 type_constrain.txt 限制的测试结果
        :type type_constrain: bool
        """

        #: 训练数据采样器
        self.sampler: typing.Union[TradSampler, RGCNSampler, CompGCNSampler] = sampler
        #: 实体的个数
        self.ent_tol: int = sampler.ent_tol
        #: valid2id.txt
        self.valid_file: str = valid_file
        #: test2id.txt
        self.test_file: str = test_file

        #: 验证集三元组的个数
        self.valid_tol: int = 0
        #: 测试集三元组的个数
        self.test_tol: int = 0

        #: 验证集三元组
        self.valid_triples: list[tuple[int, int, int]] = []
        #: 测试集三元组
        self.test_triples: list[tuple[int, int, int]] = []
        #: 知识图谱所有三元组
        self.all_true_triples: set[tuple[int, int, int]] = set()

        self.get_valid_test_triples_id()

        #: 知识图谱中所有 h-r 对对应的 t 集合
        self.hr2t_all: ddict[set] = ddict(set)
        #: 知识图谱中所有 r-t 对对应的 h 集合
        self.rt2h_all: ddict[set] = ddict(set)

        #: 是否报告 type_constrain.txt 限制的测试结果
        self.type_constrain: bool = type_constrain

        if self.type_constrain:
            construct_type_constrain(
                in_path = self.sampler.in_path,
                train_file = self.sampler.train_file,
                valid_file = self.valid_file,
                test_file = self.test_file
            )
            #: 知识图谱中所有 r 存在头实体种类
            self.rel_heads: ddict[set] = ddict(set)
            #: 知识图谱中所有 r 存在尾实体种类
            self.rel_tails: ddict[set] = ddict(set)
            self.get_type_constrain_id()

    def get_valid_test_triples_id(self):

        """读取 :py:attr:`valid_file` 文件和 :py:attr:`test_file` 文件。"""
                
        with open(os.path.join(self.sampler.in_path, self.valid_file)) as f:
            self.valid_tol = (int)(f.readline())
            for line in f:
                h, t, r = line.strip().split()
                self.valid_triples.append((int(h), int(r), int(t)))
                
        with open(os.path.join(self.sampler.in_path, self.test_file)) as f:
            self.test_tol = (int)(f.readline())
            for line in f:
                h, t, r = line.strip().split()
                self.test_triples.append((int(h), int(r), int(t)))

        self.all_true_triples = set(
            self.sampler.train_triples + self.valid_triples + self.test_triples
        )
        
    def get_type_constrain_id(self):

        """读取 type_constrain.txt 文件。"""
                
        with open(os.path.join(self.sampler.in_path, "type_constrain.txt")) as f:
            rel_tol = (int)(f.readline())
            first_line = True
            for line in f:
                rel_types = line.strip().split("\t")
                for entity in rel_types[2:]:
                    if first_line:
                        self.rel_heads[int(rel_types[0])].add(int(entity))
                    else:
                        self.rel_tails[int(rel_types[0])].add(int(entity))
                first_line = not first_line

        for rel in self.rel_heads:
            self.rel_heads[rel] = torch.tensor(list(self.rel_heads[rel]))
        for rel in self.rel_tails:
            self.rel_tails[rel] = torch.tensor(list(self.rel_tails[rel]))

    def get_hr2t_rt2h_from_all(self):

        """获得 :py:attr:`hr2t_all` 和 :py:attr:`rt2h_all` 。"""

        for h, r, t in self.all_true_triples:
            self.hr2t_all[(h, r)].add(t)
            self.rt2h_all[(r, t)].add(h)
        for h, r in self.hr2t_all:
            self.hr2t_all[(h, r)] = torch.tensor(list(self.hr2t_all[(h, r)]))
        for r, t in self.rt2h_all:
            self.rt2h_all[(r, t)] = torch.tensor(list(self.rt2h_all[(r, t)]))

    def sampling(
        self,
        data: list[tuple[int, int, int]]) -> dict[str, torch.Tensor]:

        """采样函数。该方法未实现，子类必须重写该方法，否则抛出 :py:class:`NotImplementedError` 错误。
        
        :param data: 测试的正确三元组
        :type data: list[tuple[int, int, int]]
        :returns: 测试数据
        :rtype: dict[str, torch.Tensor]
        """
        
        raise NotImplementedError

    def get_valid(self) -> list[tuple[int, int, int]]:

        """
        返回验证集三元组。

        :returns: :py:attr:`valid_triples`
        :rtype: list[tuple[int, int, int]]
        """

        return self.valid_triples

    def get_test(self) -> list[tuple[int, int, int]]:

        """
        返回测试集三元组。

        :returns: :py:attr:`test_triples`
        :rtype: list[tuple[int, int, int]]
        """

        return self.test_triples

    def get_all_true_triples(self) -> set[tuple[int, int, int]]:

        """
        返回知识图谱所有三元组。

        :returns: :py:attr:`all_true_triples`
        :rtype: set[tuple[int, int, int]]
        """

        return self.all_true_triples 


# coding:utf-8
#
# unike/module/strategy/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 28, 2023
# 
# 该头文件定义了 strategy 接口.

"""训练策略部分。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .Strategy import Strategy
from .NegativeSampling import NegativeSampling, get_negative_sampling_hpo_config
from .RGCNSampling import RGCNSampling, get_rgcn_sampling_hpo_config
from .CompGCNSampling import CompGCNSampling, get_compgcn_sampling_hpo_config

__all__ = [
    'Strategy',
    'NegativeSampling',
    'get_negative_sampling_hpo_config',
    'RGCNSampling',
    'get_rgcn_sampling_hpo_config',
    'CompGCNSampling',
    'get_compgcn_sampling_hpo_config'
]


# coding:utf-8
#
# unike/module/model/TransH.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2024
# 
# 该头文件定义了 TransH.

"""
TransH - 是第二个平移模型，将关系建模为超平面上的平移操作。
"""

import torch
import typing
import torch.nn as nn
import torch.nn.functional as F
from .Model import Model
from typing_extensions import override

class TransH(Model):

	"""
	``TransH`` :cite:`TransH` 提出于 2014 年，是第二个平移模型，将关系建模为超平面上的平移操作。
	
	评分函数为:
	
	.. math::

		\Vert (h-r_w^T hr_w)+r_d-(t-r_w^T tr_w)\Vert_{L_1/L_2}

	正三元组的评分函数的值越小越好，如果想获得更详细的信息请访问 :ref:`TransH <transh>`。

	例子::

		from unike.data import KGEDataLoader, BernSampler, TradTestSampler
		from unike.module.model import TransH
		from unike.module.loss import MarginLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = "../../benchmarks/FB15K237/",
			batch_size = 4096,
			neg_ent = 25,
			test = True,
			test_batch_size = 30, 
			num_workers = 16,
			train_sampler = BernSampler,
			test_sampler = TradTestSampler
		)
		
		# define the model
		transh = TransH(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = 200, 
			p_norm = 1, 
			norm_flag = True)
		
		# define the loss function
		model = NegativeSampling(
			model = transh, 
			loss = MarginLoss(margin = 4.0),
			# regul_rate = 0.01
		)
		
		# test the model
		tester = Tester(model = transh, data_loader = dataloader, use_gpu = True, device = 'cuda:1')
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
			epochs = 1000, lr = 0.5, use_gpu = True, device = 'cuda:1',
			tester = tester, test = True, valid_interval = 100,
			log_interval = 100, save_interval = 100, save_path = '../../checkpoint/transh.pth',
			delta = 0.01)
		trainer.run()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim: int = 100,
		p_norm: int = 1,
		norm_flag: bool = True,
		margin: float | None = None):
		
		"""创建 TransH 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim: 实体、关系嵌入向量和和法向量的维度
		:type dim: int
		:param p_norm: 评分函数的距离函数, 按照原论文，这里可以取 1 或 2。
		:type p_norm: int
		:param norm_flag: 是否利用 :py:func:`torch.nn.functional.normalize` 对实体和关系嵌入的最后一维执行 L2-norm。
		:type norm_flag: bool
		:param margin: 当使用 ``RotatE`` :cite:`RotatE` 的损失函数 :py:class:`unike.module.loss.SigmoidLoss`，需要提供此参数，将 ``TransE`` :cite:`TransE` 的正三元组的评分由越小越好转化为越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。
		:type margin: float
		"""

		super(TransH, self).__init__(ent_tol, rel_tol)
		
		#: 实体、关系嵌入向量和和法向量的维度
		self.dim: int = dim
		#: 评分函数的距离函数, 按照原论文，这里可以取 1 或 2。
		self.p_norm: int = p_norm
		#: 是否利用 :py:func:`torch.nn.functional.normalize` 
		#: 对实体和关系嵌入向量的最后一维执行 L2-norm。
		self.norm_flag: bool = norm_flag
		
		#: 根据实体个数，创建的实体嵌入
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim)
		#: 根据关系个数，创建的关系嵌入
		self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim)
		#: 根据关系个数，创建的法向量
		self.norm_vector: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim)

		if margin != None:
			#: 当使用 ``RotatE`` :cite:`RotatE` 的损失函数 :py:class:`unike.module.loss.SigmoidLoss`，需要提供此参数，将 ``TransE`` :cite:`TransE` 的正三元组的评分由越小越好转化为越大越好，如果想获得更详细的信息请访问 :ref:`RotatE <rotate>`。
			self.margin: torch.nn.parameter.Parameter = nn.Parameter(torch.Tensor([margin]))
			self.margin.requires_grad = False
			self.margin_flag = True
		else:
			self.margin_flag = False

		nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
		nn.init.xavier_uniform_(self.rel_embeddings.weight.data)
		nn.init.xavier_uniform_(self.norm_vector.weight.data)

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
		norm_vector = self.norm_vector(triples[:, 1]).unsqueeze(dim=1)
		head_emb = self._transfer(head_emb, norm_vector)
		tail_emb = self._transfer(tail_emb, norm_vector)
		score = self._calc(head_emb, relation_emb, tail_emb)

		if self.margin_flag:
			return self.margin - score
		else:
			return score

	def _transfer(
		self,
		e: torch.Tensor,
		norm: torch.Tensor) -> torch.Tensor:

		"""
		将头实体或尾实体的向量投影到超平面上。
		
		:param e: 头实体或尾实体向量。
		:type e: torch.Tensor
		:param norm: 法向量
		:type norm: torch.Tensor
		:returns: 投影后的实体向量
		:rtype: torch.Tensor
		"""

		norm = F.normalize(norm, p = 2, dim = -1)
		return e - torch.sum(e * norm, -1, True) * norm

	def _calc(
		self,
		h: torch.Tensor,
		r: torch.Tensor,
		t: torch.Tensor) -> torch.Tensor:

		"""计算 TransH 的评分函数。
		
		:param h: 头实体的向量。
		:type h: torch.Tensor
		:param r: 关系实体的向量。
		:type r: torch.Tensor
		:param t: 尾实体的向量。
		:type t: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		# 对嵌入的最后一维进行归一化
		if self.norm_flag:
			h = F.normalize(h, 2, -1)
			r = F.normalize(r, 2, -1)
			t = F.normalize(t, 2, -1)
		
		score = (h + r) - t
		
		# 利用距离函数计算得分
		score = torch.norm(score, self.p_norm, -1)
		return score

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode: str) -> torch.Tensor:
		
		"""TransH 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:param mode: 'head_predict' 或 'tail_predict'
		:type mode: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
		norm_vector = self.norm_vector(triples[:, 1]).unsqueeze(dim=1)
		head_emb = self._transfer(head_emb, norm_vector)
		tail_emb = self._transfer(tail_emb, norm_vector)
		score = self._calc(head_emb, relation_emb, tail_emb)
		
		if self.margin_flag:
			score = self.margin - score
			return score
		else:
			return -score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_head_emb, pos_relation_emb, pos_tail_emb = self.tri2emb(pos_sample)
		pos_norm_vector = self.norm_vector(pos_sample[:, 1]).unsqueeze(dim=1)
		if mode == "bern":
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(neg_sample)
		else:
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)
		neg_norm_vector = self.norm_vector(pos_sample[:, 1]).unsqueeze(dim=1)

		pos_regul = (torch.mean(pos_head_emb ** 2) + 
					 torch.mean(pos_relation_emb ** 2) + 
					 torch.mean(pos_tail_emb ** 2) +
					 torch.mean(pos_norm_vector ** 2)) / 4

		neg_regul = (torch.mean(neg_head_emb ** 2) + 
					 torch.mean(neg_relation_emb ** 2) + 
					 torch.mean(neg_tail_emb ** 2) +
					 torch.mean(neg_norm_vector ** 2)) / 4

		regul = (pos_regul + neg_regul) / 2
		
		return regul

def get_transh_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`TransH` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'TransH'
			},
			'dim': {
				'values': [50, 100, 200]
			},
			'p_norm': {
				'values': [1, 2]
			},
			'norm_flag': {
				'value': True
			}
		}

	:returns: :py:class:`TransH` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'TransH'
		},
		'dim': {
			'values': [50, 100, 200]
		},
		'p_norm': {
			'values': [1, 2]
		},
		'norm_flag': {
			'value': True
		}
	}
		
	return parameters_dict



# coding:utf-8
#
# unike/data/RGCNSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 21, 2024
#
# R-GCN 的数据采样器.

"""
RGCNSampler - R-GCN 的数据采样器。
"""

import dgl
import torch
import typing
import warnings
import numpy as np
from .RevSampler import RevSampler

warnings.filterwarnings("ignore")

class RGCNSampler(RevSampler):

    """``R-GCN`` :cite:`R-GCN` 的训练数据采样器。

    例子::

        from unike.data import RGCNSampler, CompGCNSampler
        from torch.utils.data import DataLoader

        #: 训练数据采样器
        train_sampler: typing.Union[typing.Type[RGCNSampler], typing.Type[CompGCNSampler]] = train_sampler(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file,
            batch_size=batch_size,
            neg_ent=neg_ent
        )

        #: 训练集三元组
        data_train: list[tuple[int, int, int]] = train_sampler.get_train()

        train_dataloader = DataLoader(
            data_train,
            shuffle=True,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=True,
            drop_last=True,
            collate_fn=train_sampler.sampling,
        )
    """
    
    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt",
        batch_size: int | None = None,
        neg_ent: int = 1):

        """创建 RGCNSampler 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        :param batch_size: batch size
        :type batch_size: int | None
        :param neg_ent: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity (head + tail)
        :type neg_ent: int
        """

        super().__init__(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file
        )

        #: batch size
        self.batch_size: int = batch_size
        #: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity (head + tail)
        self.neg_ent: int = neg_ent

        self.entity   = None
        self.triples  = None
        self.label    = None
        self.graph    = None
        self.relation = None
        self.norm     = None

    def sampling(
        self,
        pos_triples: list[tuple[int, int, int]]) -> dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]:

        """``R-GCN`` :cite:`R-GCN` 的采样函数。
        
        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[int, int, int]]
        :returns: ``R-GCN`` :cite:`R-GCN` 的训练数据
        :rtype: dict[str, typing.Union[dgl.DGLGraph , torch.Tensor]]
        """
        
        batch_data = {}
        
        pos_triples = np.array(pos_triples)
        pos_triples, self.entity = self.sampling_positive(pos_triples)
        head_triples = self.sampling_negative('head', pos_triples)
        tail_triples = self.sampling_negative('tail', pos_triples)
        self.triples = np.concatenate((pos_triples,head_triples,tail_triples))
        batch_data['entity']  = self.entity
        batch_data['triples'] = torch.from_numpy(self.triples)
        
        self.label = torch.zeros((len(self.triples),1))
        self.label[0 : self.batch_size] = 1
        batch_data['label'] = self.label
        
        split_size = int(self.batch_size * 0.5) 
        graph_split_ids = np.random.choice(
            self.batch_size,
            size=split_size, 
            replace=False
        )
        head,rela,tail = pos_triples.transpose()
        head = torch.tensor(head[graph_split_ids], dtype=torch.long).contiguous()
        rela = torch.tensor(rela[graph_split_ids], dtype=torch.long).contiguous()
        tail = torch.tensor(tail[graph_split_ids], dtype=torch.long).contiguous()
        self.graph, self.relation, self.norm = self.build_graph(len(self.entity), (head,rela,tail), -1)
        batch_data['graph']    = self.graph
        batch_data['relation'] = self.relation
        batch_data['norm']     = self.norm

        return batch_data

    def sampling_positive(
        self,
        positive_triples: list[tuple[int, int, int]]) -> tuple[np.ndarray, torch.Tensor]:

        """为创建子图重新采样三元组子集，重排实体 ID。
        
        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[int, int, int]]
        :returns: 三元组子集和原始的实体 ID
        :rtype: tuple[numpy.ndarray, torch.Tensor]
        """

        edges = np.random.choice(
            np.arange(len(positive_triples)),
            size = self.batch_size,
            replace=False
        )
        edges = positive_triples[edges]
        head, rela, tail = np.array(edges).transpose()
        entity, index = np.unique((head, tail), return_inverse=True) 
        head, tail = np.reshape(index, (2, -1))

        return np.stack((head,rela,tail)).transpose(), \
                torch.from_numpy(entity).view(-1,1).long()

    def sampling_negative(
        self,
        mode: int,
        pos_triples: list[tuple[int, int, int]]) -> np.ndarray:

        """采样负三元组。

        :param mode: 'head' 或 'tail'
        :type mode: str
        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[int, int, int]]
        :returns: 负三元组
        :rtype: numpy.ndarray
        """

        neg_random = np.random.choice(
            len(self.entity), 
            size = self.neg_ent * len(pos_triples)
        )
        neg_samples = np.tile(pos_triples, (self.neg_ent, 1))
        if mode == 'head':
            neg_samples[:,0] = neg_random
        elif mode == 'tail':
            neg_samples[:,2] = neg_random
        return neg_samples

    def build_graph(
        self,
        num_ent: int,
        triples: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        power: int = -1) -> tuple[dgl.DGLGraph, torch.Tensor, torch.Tensor]:

        """建立子图。

        :param num_ent: 子图的节点数
        :type num_ent: int
        :param triples: 知识图谱中的正确三元组子集
        :type triples: tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        :param power: 幂
        :type power: int
        :returns: 子图、关系、边的归一化系数
        :rtype: tuple[dgl.DGLGraph, torch.Tensor, torch.Tensor]
        """

        head, rela, tail = triples[0], triples[1], triples[2]
        graph = dgl.graph(([], []))
        graph.add_nodes(num_ent)
        graph.add_edges(head, tail)
        node_norm = self.comp_deg_norm(graph, power)
        edge_norm = self.node_norm_to_edge_norm(graph,node_norm)
        rela = torch.tensor(rela)
        return graph, rela, edge_norm

    def comp_deg_norm(
        self,
        graph: dgl.DGLGraph,
        power: int = -1) -> torch.Tensor:

        """根据目标节点度计算目标节点的归一化系数。

        :param graph: 子图
        :type graph: dgl.DGLGraph
        :param power: 幂
        :type power: int
        :returns: 节点的归一化系数
        :rtype: torch.Tensor
        """

        graph = graph.local_var()
        in_deg = graph.in_degrees(range(graph.number_of_nodes())).float().numpy()
        norm = in_deg.__pow__(power)
        norm[np.isinf(norm)] = 0
        return torch.from_numpy(norm)

    def node_norm_to_edge_norm(
        self,
        graph: dgl.DGLGraph,
        node_norm: torch.Tensor) -> torch.Tensor:

        """根据目标节点度计算每条边的归一化系数。

        :param graph: 子图
        :type graph: dgl.DGLGraph
        :param node_norm: 节点的归一化系数
        :type node_norm: torch.Tensor
        :returns: 边的归一化系数
        :rtype: torch.Tensor
        """
        
        graph = graph.local_var()
        # convert to edge norm
        graph.ndata['norm'] = node_norm.view(-1,1)
        graph.apply_edges(lambda edges : {'norm' : edges.dst['norm']})
        return graph.edata['norm']


# coding:utf-8
#
# unike/module/model/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 28, 2023
# 
# 该头文件定义了 model 接口.

"""KGE 模型部分。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .Model import Model
from .TransE import TransE, get_transe_hpo_config
from .TransH import TransH, get_transh_hpo_config
from .TransR import TransR, get_transr_hpo_config
from .TransD import TransD, get_transd_hpo_config
from .RotatE import RotatE, get_rotate_hpo_config
from .RESCAL import RESCAL, get_rescal_hpo_config
from .DistMult import DistMult, get_distmult_hpo_config
from .HolE import HolE, get_hole_hpo_config
from .ComplEx import ComplEx, get_complex_hpo_config
from .Analogy import Analogy, get_analogy_hpo_config
from .SimplE import SimplE, get_simple_hpo_config
from .RGCN import RGCN, get_rgcn_hpo_config
from .CompGCN import CompGCN, CompGCNCov, get_compgcn_hpo_config

__all__ = [
    'Model',
    'TransE',
    'get_transe_hpo_config',
    'TransH',
    'get_transh_hpo_config',
    'TransR',
    'get_transr_hpo_config',
    'TransD',
    'get_transd_hpo_config',
    'RotatE',
    'get_rotate_hpo_config',
    'RESCAL',
    'get_rescal_hpo_config',
    'DistMult',
    'get_distmult_hpo_config',
    'HolE',
    'get_hole_hpo_config',
    'ComplEx',
    'get_complex_hpo_config',
    'Analogy',
    'get_analogy_hpo_config',
    'SimplE',
    'get_simple_hpo_config',
    'RGCN',
    'get_rgcn_hpo_config',
    'CompGCN',
    'CompGCNCov',
    'get_compgcn_hpo_config'
]


# coding:utf-8
#
# unike/module/strategy/Strategy.py
#
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 4, 2023
#
# 该脚本定义了损失函数的基类.

"""
Strategy - 该脚本定义了训练策略的基类。
"""

from ..BaseModule import BaseModule

class Strategy(BaseModule):

	"""
	继承自 :py:class:`unike.module.BaseModule`，什么额外的属性都没有增加。
	"""

	def __init__(self):
		
		"""创建 Loss 对象。"""
		
		super(Strategy, self).__init__()


# coding:utf-8
#
# unike/data/RGCNTestSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 29, 2024
#
# R-GCN 的测试数据采样器.

"""
RGCNTestSampler - R-GCN 的测试数据采样器。
"""

import os
import dgl
import torch
import typing
import numpy as np
from .TestSampler import TestSampler
from .RGCNSampler import RGCNSampler
from .CompGCNSampler import CompGCNSampler
from typing_extensions import override

class RGCNTestSampler(TestSampler):

    """``R-GCN`` :cite:`R-GCN` 的测试数据采样器。

    例子::

        from unike.data import RGCNTestSampler, CompGCNTestSampler
        from torch.utils.data import DataLoader

        #: 测试数据采样器
        test_sampler: typing.Union[typing.Type[RGCNTestSampler], typing.Type[CompGCNTestSampler]] = test_sampler(
            sampler=train_sampler,
            valid_file=valid_file,
            test_file=test_file,
        )
    
        #: 验证集三元组
        data_val: list[tuple[int, int, int]] = test_sampler.get_valid()
        #: 测试集三元组
        data_test: list[tuple[int, int, int]] = test_sampler.get_test()

        val_dataloader = DataLoader(
            data_val,
            shuffle=False,
            batch_size=test_batch_size,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=test_sampler.sampling,
        )

        test_dataloader = DataLoader(
            data_test,
            shuffle=False,
            batch_size=test_batch_size,
            num_workers=num_workers,
            pin_memory=True,
            collate_fn=test_sampler.sampling,
        )
    """

    def __init__(
        self,
        sampler: typing.Union[RGCNSampler, CompGCNSampler],
        valid_file: str = "valid2id.txt",
        test_file: str = "test2id.txt",
        type_constrain: bool = True):

        """创建 RGCNTestSampler 对象。

        :param sampler: 训练数据采样器。
        :type sampler: typing.Union[RGCNSampler, CompGCNSampler]
        :param valid_file: valid2id.txt
        :type valid_file: str
        :param test_file: test2id.txt
        :type test_file: str
        :param type_constrain: 是否报告 type_constrain.txt 限制的测试结果
        :type type_constrain: bool
        """

        super().__init__(
            sampler=sampler,
            valid_file=valid_file,
            test_file=test_file,
            type_constrain = type_constrain
        )

        #: 训练集三元组
        self.triples: list[tuple[int, int, int]] = self.sampler.t_triples if isinstance(self.sampler, CompGCNSampler) else self.sampler.train_triples
        #: 幂
        self.power: float = -1

        self.add_valid_test_reverse_triples()
        self.get_hr2t_rt2h_from_all()

    @override
    def get_valid_test_triples_id(self):

        """读取 :py:attr:`valid_file` 文件和 :py:attr:`test_file` 文件。"""
                
        with open(os.path.join(self.sampler.in_path, self.valid_file)) as f:
            self.valid_tol = (int)(f.readline())
            for line in f:
                h, t, r = line.strip().split()
                self.valid_triples.append((int(h), int(r), int(t)))
                
        with open(os.path.join(self.sampler.in_path, self.test_file)) as f:
            self.test_tol = (int)(f.readline())
            for line in f:
                h, t, r = line.strip().split()
                self.test_triples.append((int(h), int(r), int(t)))

    def add_valid_test_reverse_triples(self):

        """对于每一个三元组 (h, r, t)，生成相反关系三元组 (t, r`, h): r` = r + rel_tol。"""

        tol = int(self.sampler.rel_tol / 2)
                
        with open(os.path.join(self.sampler.in_path, self.valid_file)) as f:
            f.readline()
            for line in f:
                h, t, r = line.strip().split()
                self.valid_triples.append(
                    (int(t), int(r) + tol, int(h))
                )
                
        with open(os.path.join(self.sampler.in_path, self.test_file)) as f:
            f.readline()
            for line in f:
                h, t, r = line.strip().split()
                self.test_triples.append(
                    (int(t), int(r) + tol, int(h))
                )
                
        self.all_true_triples = set(
            self.triples + self.valid_triples + self.test_triples
        )

    @override
    def get_type_constrain_id(self):

        """读取 type_constrain.txt 文件。"""

        tol = int(self.sampler.rel_tol / 2)
                
        with open(os.path.join(self.sampler.in_path, "type_constrain.txt")) as f:
            rel_tol = (int)(f.readline())
            first_line = True
            for line in f:
                rel_types = line.strip().split("\t")
                for entity in rel_types[2:]:
                    if first_line:
                        self.rel_heads[int(rel_types[0])].add(int(entity))
                        self.rel_tails[int(rel_types[0]) + tol].add(int(entity))
                    else:
                        self.rel_tails[int(rel_types[0])].add(int(entity))
                        self.rel_heads[int(rel_types[0]) + tol].add(int(entity))
                first_line = not first_line

        for rel in self.rel_heads:
            self.rel_heads[rel] = torch.tensor(list(self.rel_heads[rel]))
        for rel in self.rel_tails:
            self.rel_tails[rel] = torch.tensor(list(self.rel_tails[rel]))

    @override
    def sampling(
        self,
        data: list[tuple[int, int, int]]) -> dict[str, typing.Union[dgl.DGLGraph , torch.Tensor]]:

        """``R-GCN`` :cite:`R-GCN` 的测试数据采样函数。
        
        :param data: 测试的正确三元组
        :type data: list[tuple[int, int, int]]
        :returns: ``R-GCN`` :cite:`R-GCN` 的测试数据
        :rtype: dict[str, typing.Union[dgl.DGLGraph , torch.Tensor]]
        """
        
        batch_data = {}
        head_label = torch.zeros(len(data), self.ent_tol)
        tail_label = torch.zeros(len(data), self.ent_tol)
        for idx, triple in enumerate(data):
            head, rel, tail = triple
            head_label[idx][self.rt2h_all[(rel, tail)]] = 1.0
            tail_label[idx][self.hr2t_all[(head, rel)]] = 1.0
        if self.type_constrain:
            head_label_type = torch.ones(len(data), self.ent_tol)
            tail_laebl_type = torch.ones(len(data), self.ent_tol)
            for idx, triple in enumerate(data):
                head, rel, tail = triple
                head_label_type[idx][self.rel_heads[rel]] = 0.0
                tail_laebl_type[idx][self.rel_tails[rel]] = 0.0
                head_label_type[idx][self.rt2h_all[(rel, tail)]] = 1.0
                tail_laebl_type[idx][self.hr2t_all[(head, rel)]] = 1.0
            batch_data["head_label_type"] = head_label_type
            batch_data["tail_label_type"] = tail_laebl_type
        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["head_label"] = head_label
        batch_data["tail_label"] = tail_label
        
        graph, rela, norm = self.sampler.build_graph(self.ent_tol, np.array(self.triples).transpose(), self.power)
        batch_data["graph"]  = graph
        batch_data["rela"]   = rela
        batch_data["norm"]   = norm
        batch_data["entity"] = torch.arange(0, self.ent_tol, dtype=torch.long).view(-1,1)
        
        return batch_data


# coding:utf-8
#
# unike/module/loss/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 28, 2023
# 
# 该头文件定义了 loss 接口.

"""损失函数部分。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .Loss import Loss
from .MarginLoss import MarginLoss, get_margin_loss_hpo_config
from .SigmoidLoss import SigmoidLoss, get_sigmoid_loss_hpo_config
from .SoftplusLoss import SoftplusLoss, get_softplus_loss_hpo_config
from .RGCNLoss import RGCNLoss, get_rgcn_loss_hpo_config
from .CompGCNLoss import CompGCNLoss, get_compgcn_loss_hpo_config

__all__ = [
    'Loss',
    'MarginLoss',
    'get_margin_loss_hpo_config',
    'SigmoidLoss',
    'get_sigmoid_loss_hpo_config',
    'SoftplusLoss',
    'get_softplus_loss_hpo_config',
    'RGCNLoss',
    'get_rgcn_loss_hpo_config',
    'CompGCNLoss',
    'get_compgcn_loss_hpo_config'
]


# coding:utf-8
#
# unike/data/UniSampler.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 29, 2024
#
# 平移模型和语义匹配模型的训练集数据采样器.

"""
UniSampler - 平移模型和语义匹配模型的训练集数据采样器。
"""

import torch
import typing
import warnings
import numpy as np
from .TradSampler import TradSampler
from typing_extensions import override

warnings.filterwarnings("ignore")

class UniSampler(TradSampler):

    """平移模型和语义匹配模型的训练集普通的数据采样器（均值分布）。
    """
    
    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt",
        batch_size: int | None = None,
        neg_ent: int = 1):

        """创建 UniSampler 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        :param batch_size: batch size 在该采样器中不起作用，只是占位符。
        :type batch_size: int | None
        :param neg_ent: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity
        :type neg_ent: int
        """

        super().__init__(
            in_path=in_path,
            ent_file=ent_file,
            rel_file=rel_file,
            train_file=train_file,
            batch_size = batch_size,
            neg_ent = neg_ent
        )

        self.cross_sampling_flag = 0

    @override
    def sampling(
        self,
        pos_triples: list[tuple[int, int, int]]) -> dict[str, typing.Union[str, torch.Tensor]]:
        
        """平移模型和语义匹配模型的训练集普通的数据采样函数（均匀分布）。
        
        :param pos_triples: 知识图谱中的正确三元组
        :type pos_triples: list[tuple[int, int, int]]
        :returns: 平移模型和语义匹配模型的训练数据
        :rtype: dict[str, typing.Union[str, torch.Tensor]]
        """
        
        batch_data = {}
        neg_ent_sample = []
        self.cross_sampling_flag = 1 - self.cross_sampling_flag
        if self.cross_sampling_flag == 0:
            batch_data['mode'] = "head-batch"
            for h, r, t in pos_triples:
                neg_head = self.head_batch(t, r, self.neg_ent)
                neg_ent_sample.append(neg_head)
        else:
            batch_data['mode'] = "tail-batch"
            for h, r, t in pos_triples:
                neg_tail = self.tail_batch(h, r, self.neg_ent)
                neg_ent_sample.append(neg_tail)

        batch_data["positive_sample"] = torch.LongTensor(np.array(pos_triples))
        batch_data['negative_sample'] = torch.LongTensor(np.array(neg_ent_sample))
        return batch_data

    def head_batch(
        self,
        t: int,
        r: int,
        neg_size: int= None) -> np.ndarray:

        """替换头实体构建负三元组。

        :param t: 尾实体
        :type t: int
        :param r: 关系
        :type r: int
        :param neg_size: 负三元组个数
        :type neg_size: int
        :returns: 负三元组中的头实体列表
        :rtype: numpy.ndarray
        """
        
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_head(t, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def tail_batch(
        self,
        h: int,
        r: int,
        neg_size: int = None) -> np.ndarray:
        
        """替换尾实体构建负三元组。

        :param h: 头实体
        :type h: int
        :param r: 关系
        :type r: int
        :param neg_size: 负三元组个数
        :type neg_size: int
        :returns: 负三元组中的尾实体列表
        :rtype: numpy.ndarray
        """
        
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_tail(h, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]


# coding:utf-8
#
# unike/module/model/DistMult.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 6, 2023
# 
# 该头文件定义了 DistMult.

"""
DistMult - 最简单的双线性模型，与 TransE 参数量相同，因此非常容易的应用于大型的知识图谱。
"""

import torch
import typing
import torch.nn as nn
from .Model import Model
from typing_extensions import override

class DistMult(Model):

	"""
	``DistMult`` :cite:`DistMult` 提出于 2015 年，最简单的双线性模型，与 TransE 参数量相同，因此非常容易的应用于大型的知识图谱。

	评分函数为:

	.. math::
	
		\sum_{i=1}^{n}h_ir_it_i
	
	为逐元素多线性点积（element-wise multi-linear dot product），正三元组的评分函数的值越大越好，负三元组越小越好，如果想获得更详细的信息请访问 :ref:`DistMult <distMult>`。

	例子::

		from unike.utils import WandbLogger
		from unike.data import KGEDataLoader, BernSampler, TradTestSampler
		from unike.module.model import DistMult
		from unike.module.loss import SoftplusLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		wandb_logger = WandbLogger(
			project="unike",
			name="DistMult-WN18RR",
			config=dict(
				in_path = '../../benchmarks/WN18RR/',
				batch_size = 4096,
				neg_ent = 25,
				test = True,
				test_batch_size = 10,
				num_workers = 16,
				dim = 200,
				regul_rate = 1.0,
				use_gpu = True,
				device = 'cuda:0',
				epochs = 2000,
				lr = 0.5,
				opt_method = 'adagrad',
				valid_interval = 100,
				log_interval = 100,
				save_interval = 100,
				save_path = '../../checkpoint/distMult.pth'
			)
		)
		
		config = wandb_logger.config
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = config.in_path, 
			batch_size = config.batch_size,
			neg_ent = config.neg_ent,
			test = config.test,
			test_batch_size = config.test_batch_size,
			num_workers = config.num_workers,
			train_sampler = BernSampler,
			test_sampler = TradTestSampler
		)
		
		# define the model
		distmult = DistMult(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = config.dim
		)
		
		# define the loss function
		model = NegativeSampling(
			model = distmult, 
			loss = SoftplusLoss(), 
			regul_rate = config.regul_rate
		)
		
		# test the model
		tester = Tester(model = distmult, data_loader = dataloader, use_gpu = config.use_gpu, device = config.device)
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(), epochs = config.epochs,
			lr = config.lr, opt_method = config.opt_method, use_gpu = config.use_gpu, device = config.device,
			tester = tester, test = config.test, valid_interval = config.valid_interval,
			log_interval = config.log_interval, save_interval = config.save_interval,
			save_path = config.save_path, wandb_logger = wandb_logger)
		trainer.run()
		
		# close your wandb run
		wandb_logger.finish()
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int,
		dim: int = 100):

		"""创建 DistMult 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		:param dim: 实体嵌入向量和关系对角矩阵的维度
		:type dim: int
		"""

		super(DistMult, self).__init__(ent_tol, rel_tol)

		#: 实体嵌入向量和关系对角矩阵的维度
		self.dim: int = dim
		#: 根据实体个数，创建的实体嵌入
		self.ent_embeddings: torch.nn.Embedding = nn.Embedding(self.ent_tol, self.dim)
		#: 根据关系个数，创建的关系对角矩阵
		self.rel_embeddings: torch.nn.Embedding = nn.Embedding(self.rel_tol, self.dim)

		nn.init.xavier_uniform_(self.ent_embeddings.weight.data)
		nn.init.xavier_uniform_(self.rel_embeddings.weight.data)

	@override
	def forward(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> torch.Tensor:

		"""
		定义每次调用时执行的计算。
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
		score = self._calc(head_emb, relation_emb, tail_emb)
		return score

	def _calc(
		self,
		h: torch.Tensor,
		r: torch.Tensor,
		t: torch.Tensor) -> torch.Tensor:

		"""计算 DistMult 的评分函数。
		
		:param h: 头实体的向量。
		:type h: torch.Tensor
		:param r: 关系的对角矩阵。
		:type r: torch.Tensor
		:param t: 尾实体的向量。
		:type t: torch.Tensor
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		score = (h * r) * t

		# 计算得分
		score = torch.sum(score, -1)
		return score

	@override
	def predict(
		self,
		data: dict[str, typing.Union[torch.Tensor,str]],
		mode) -> torch.Tensor:
		
		"""DistMult 的推理方法。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor,str]]
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		triples = data["positive_sample"]
		head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
		score = self._calc(head_emb, relation_emb, tail_emb)
		return score

	def regularization(
		self,
		data: dict[str, typing.Union[torch.Tensor, str]]) -> torch.Tensor:

		"""L2 正则化函数（又称权重衰减），在损失函数中用到。
		
		:param data: 数据。
		:type data: dict[str, typing.Union[torch.Tensor, str]]
		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		pos_sample = data["positive_sample"]
		neg_sample = data["negative_sample"]
		mode = data["mode"]
		pos_head_emb, pos_relation_emb, pos_tail_emb = self.tri2emb(pos_sample)
		if mode == "bern":
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(neg_sample)
		else:
			neg_head_emb, neg_relation_emb, neg_tail_emb = self.tri2emb(pos_sample, neg_sample, mode)

		pos_regul = (torch.mean(pos_head_emb ** 2) + 
					 torch.mean(pos_relation_emb ** 2) + 
					 torch.mean(pos_tail_emb ** 2)) / 3

		neg_regul = (torch.mean(neg_head_emb ** 2) + 
					 torch.mean(neg_relation_emb ** 2) + 
					 torch.mean(neg_tail_emb ** 2)) / 3

		regul = (pos_regul + neg_regul) / 2

		return regul

	def l3_regularization(self):

		"""L3 正则化函数，在损失函数中用到。

		:returns: 模型参数的正则损失
		:rtype: torch.Tensor
		"""

		return (self.ent_embeddings.weight.norm(p = 3)**3 + self.rel_embeddings.weight.norm(p = 3)**3)

def get_distmult_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`DistMult` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'model': {
				'value': 'DistMult'
			},
			'dim': {
				'values': [50, 100, 200]
			}
		}

	:returns: :py:class:`DistMult` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]
	"""

	parameters_dict = {
		'model': {
			'value': 'DistMult'
		},
		'dim': {
			'values': [50, 100, 200]
		}
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/module/model/Model.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 25, 2023
# 
# 该头文件定义了 Model.

"""Model 类 - 所有 KGE 模型的基类"""

import torch
from ..BaseModule import BaseModule

class Model(BaseModule):

	"""
	继承自 :py:class:`unike.module.BaseModule`，仅仅增加了两个属性：:py:attr:`ent_tol` 和 :py:attr:`rel_tol`。
	"""

	def __init__(
		self,
		ent_tol: int,
		rel_tol: int):

		"""创建 Model 对象。

		:param ent_tol: 实体的个数
		:type ent_tol: int
		:param rel_tol: 关系的个数
		:type rel_tol: int
		"""

		super(Model, self).__init__()

		#: 实体的种类
		self.ent_tol: int = ent_tol
		#: 关系的种类
		self.rel_tol: int = rel_tol

	def forward(self) -> torch.Tensor:

		"""
		定义每次调用时执行的计算。该方法未实现，子类必须重写该方法，否则抛出 :py:class:`NotImplementedError` 错误。
		
		:py:class:`torch.nn.Module` 子类必须重写 :py:meth:`torch.nn.Module.forward`。

		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		raise NotImplementedError
	
	def predict(self) -> torch.Tensor:

		"""
		KGE 模型的推理方法。该方法未实现，子类必须重写该方法，否则抛出 :py:class:`NotImplementedError` 错误。
		
		:returns: 三元组的得分
		:rtype: torch.Tensor
		"""

		raise NotImplementedError

	def tri2emb(
		self,
		triples: torch.Tensor,
		negs: torch.Tensor = None,
		mode: str = 'single') -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

		"""
		返回三元组对应的嵌入向量。
		
		:param triples: 正确的三元组
		:type triples: torch.Tensor
		:param negs: 负三元组类别
		:type negs: torch.Tensor
		:param mode: 模式
		:type triples: str
		:returns: 头实体、关系和尾实体的嵌入向量
		:rtype: tuple[torch.Tensor, torch.Tensor, torch.Tensor]
		"""
		
		if mode == "single":
			head_emb = self.ent_embeddings(triples[:, 0]).unsqueeze(1)
			relation_emb = self.rel_embeddings(triples[:, 1]).unsqueeze(1)
			tail_emb = self.ent_embeddings(triples[:, 2]).unsqueeze(1)
			
		elif mode == "head-batch" or mode == "head_predict":
			if negs is None:
				head_emb = self.ent_embeddings.weight.data.unsqueeze(0)
			else:
				head_emb = self.ent_embeddings(negs)
				
			relation_emb = self.rel_embeddings(triples[:, 1]).unsqueeze(1)
			tail_emb = self.ent_embeddings(triples[:, 2]).unsqueeze(1)
			
		elif mode == "tail-batch" or mode == "tail_predict": 
			head_emb = self.ent_embeddings(triples[:, 0]).unsqueeze(1)
			relation_emb = self.rel_embeddings(triples[:, 1]).unsqueeze(1)
			
			if negs is None:
				tail_emb = self.ent_embeddings.weight.data.unsqueeze(0)
			else:
				tail_emb = self.ent_embeddings(negs)
		
		return head_emb, relation_emb, tail_emb


# coding:utf-8
#
# unike/config/Trainer.py
#
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 9, 2024
#
# 该脚本定义了训练循环基类类.

"""
Trainer - 训练循环类。
"""

import os
import dgl
import wandb
import typing
import torch
from .Tester import Tester
import torch.optim as optim
from ..utils.Timer import Timer
from ..module.model import Model
from ..utils import WandbLogger
from torch.utils.data import DataLoader
from ..utils.EarlyStopping import EarlyStopping
from ..module.strategy import Strategy
from accelerate import Accelerator, InitProcessGroupKwargs
from datetime import timedelta
from loguru import logger


class Trainer(object):

	"""
	主要用于 KGE 模型的训练。

	例子::

		from unike.data import KGEDataLoader, BernSampler, TradTestSampler
		from unike.module.model import TransE
		from unike.module.loss import MarginLoss
		from unike.module.strategy import NegativeSampling
		from unike.config import Trainer, Tester
		
		# dataloader for training
		dataloader = KGEDataLoader(
			in_path = "../../benchmarks/FB15K/", 
			batch_size = 8192,
			neg_ent = 25,
			test = True,
			test_batch_size = 256,
			num_workers = 16,
			train_sampler = BernSampler,
			test_sampler = TradTestSampler
		)
		
		# define the model
		transe = TransE(
			ent_tol = dataloader.get_ent_tol(),
			rel_tol = dataloader.get_rel_tol(),
			dim = 50, 
			p_norm = 1, 
			norm_flag = True)
		
		# define the loss function
		model = NegativeSampling(
			model = transe, 
			loss = MarginLoss(margin = 1.0),
			regul_rate = 0.01
		)
			
		# test the model
		tester = Tester(model = transe, data_loader = dataloader, use_gpu = True, device = 'cuda:1')
		
		# train the model
		trainer = Trainer(model = model, data_loader = dataloader.train_dataloader(),
			epochs = 1000, lr = 0.01, use_gpu = True, device = 'cuda:1',
			tester = tester, test = True, valid_interval = 100,
			log_interval = 100, save_interval = 100,
			save_path = '../../checkpoint/transe.pth', delta = 0.01)
		trainer.run()
	"""

	def __init__(
		self,
		model: Strategy,
		data_loader: DataLoader,
		epochs: int = 1000,
		lr: float = 0.5,
		opt_method: str = "Adam",
		use_accelerator: bool = False,
		nccl_timeout_seconds: int = 10800,
		use_gpu: bool = True,
		device: str = "cuda:0",
		tester: Tester | None = None,
		test: bool = False,
		valid_interval: int | None = None,
		log_interval: int | None = None,
		save_interval: int | None = None,
		save_path: str | None = None,
		use_early_stopping: bool = True,
		metric: str = 'hits@10',
		patience: int = 2,
		delta: float = 0,
		wandb_logger: WandbLogger | None = None):

		"""创建 Trainer 对象。

		:param model: 包装 KGE 模型的训练策略类
		:type model: :py:class:`unike.module.strategy.Strategy`
		:param data_loader: :py:class:`torch.utils.data.DataLoader`
		:type data_loader: torch.utils.data.DataLoader
		:param epochs: 训练轮次数
		:type epochs: int
		:param lr: 学习率
		:type lr: float
		:param opt_method: 优化器: **'Adam'** or **'adam'**, **'Adagrad'** or **'adagrad'**, **'SGD'** or **'sgd'**
		:type opt_method: str
		:param use_accelerator: 使用 accelerate 进行分布式训练
		:type use_accelerator: bool
		:param nccl_timeout_seconds: nccl 超时可等待时间（秒）。默认值：10800 
		:type nccl_timeout_seconds: int
		:param use_gpu: 是否使用 gpu
		:type use_gpu: bool
		:param device: 使用哪个 gpu
		:type device: str
		:param tester: 用于模型评估的验证模型类
		:type tester: :py:class:`unike.config.Tester`
		:param test: 是否在测试集上评估模型, :py:attr:`tester` 不为空
		:type test: bool
		:param valid_interval: 训练几轮在验证集上评估一次模型, :py:attr:`tester` 不为空
		:type valid_interval: int
		:param log_interval: 训练几轮输出一次日志
		:type log_interval: int
		:param save_interval: 训练几轮保存一次模型
		:type save_interval: int
		:param save_path: 模型保存的路径
		:type save_path: str
		:param use_early_stopping: 是否启用早停，需要 :py:attr:`tester` 和 :py:attr:`save_path` 不为空
		:type use_early_stopping: bool
		:param metric: 早停使用的验证指标，可选值：**'mr'**, **'mrr'**, **'hits@N'**, **'mr_type'**, **'mrr_type'**, **'hits@N_type'**。默认值：**'hits@10'**
		:type metric: str
		:param patience: :py:attr:`unike.utils.EarlyStopping.patience` 参数，上次验证得分改善后等待多长时间。默认值：2
		:type patience: int
		:param delta: :py:attr:`unike.utils.EarlyStopping.delta` 参数，监测数量的最小变化才符合改进条件。默认值：0
		:type delta: float
		:param wandb_logger: :py:class:`unike.utils.WandbLogger` 对象
		:type wandb_logger: :py:class:`unike.utils.WandbLogger`
		"""
		
		#: 包装 KGE 模型的训练策略类，即 :py:class:`unike.module.strategy.Strategy`
		self.model: Strategy = model

		#: :py:meth:`__init__` 传入的 :py:class:`torch.utils.data.DataLoader`
		self.data_loader: torch.utils.data.DataLoader = data_loader
		#: epochs
		self.epochs: int = epochs

		#: 学习率
		self.lr: float = lr
		#: 用户传入的优化器名字字符串
		self.opt_method: str = opt_method
		#: 根据 :py:meth:`__init__` 的 ``opt_method`` 生成对应的优化器
		self.optimizer: torch.optim.SGD | torch.optim.Adagrad | torch.optim.Adam | None = None
		#: 学习率调度器
		self.scheduler: torch.optim.lr_scheduler.MultiStepLR | None = None

		#: 是否使用 gpu
		self.use_gpu: bool = use_gpu
		#: gpu，利用 ``device`` 构造的 :py:class:`torch.device` 对象
		self.device: typing.Union[torch.device, str] = torch.device(device) if self.use_gpu else "cpu"

		#: 用于模型评估的验证模型类
		self.tester: Tester | None = tester
		#: 是否在测试集上评估模型, :py:attr:`tester` 不为空
		self.test: bool = test
		#: 训练几轮在验证集上评估一次模型, :py:attr:`tester` 不为空
		self.valid_interval: int | None = valid_interval

		#: 训练几轮输出一次日志
		self.log_interval: int | None = log_interval
		#: 训练几轮保存一次模型
		self.save_interval: int | None = save_interval
		#: 模型保存的路径
		self.save_path: str | None = save_path

		#: 是否启用早停，需要 :py:attr:`tester` 和 :py:attr:`save_path` 不为空
		self.use_early_stopping: bool = use_early_stopping
		#: 早停使用的验证指标，可选值：**'mr'**, **'mrr'**, **'hits@N'**, **'mr_type'**, **'mrr_type'**, **'hits@N_type'**。默认值：**'hits@10'**
		self.metric: str = metric
		#: :py:attr:`unike.utils.EarlyStopping.patience` 参数，上次验证得分改善后等待多长时间。默认值：2
		self.patience: int = patience
		#: :py:attr:`unike.utils.EarlyStopping.delta` 参数，监测数量的最小变化才符合改进条件。默认值：0
		self.delta: float = delta
		#: 早停对象
		self.early_stopping = None

		#: :py:class:`unike.utils.WandbLogger` 对象
		self.wandb_logger = wandb_logger
  
		#: :py:class:`accelerate.Accelerator` 对象
		self.accelerator = None

		if use_accelerator:
			init_process_group_kwargs  = InitProcessGroupKwargs(timeout=timedelta(seconds=nccl_timeout_seconds))
			if self.wandb_logger:
				self.accelerator = Accelerator(log_with=self.wandb_logger.endpoint, kwargs_handlers=[init_process_group_kwargs])
				self.accelerator.init_trackers(
					project_name=self.wandb_logger.project,
					config=self.wandb_logger.config.__dict__,
					init_kwargs={
						self.wandb_logger.endpoint: {
							'name': self.wandb_logger.name
						}
					}
				)
				self.wandb_logger.logger = self.accelerator
			else:
				self.accelerator = Accelerator(kwargs_handlers=[init_process_group_kwargs])
			
			self.data_loader, self.model = self.accelerator.prepare(self.data_loader, self.model)
		else:
			if self.wandb_logger and self.wandb_logger.logger is None:
				self.wandb_logger._init()

	def configure_optimizers(self):

		"""可以通过重新实现该方法自定义配置优化器。"""

		if self.opt_method == "Adam" or self.opt_method == "adam":
			self.optimizer = optim.Adam(
				self.model.parameters(),
				lr=self.lr,
			)
		elif self.opt_method == "Adagrad" or self.opt_method == "adagrad":
			self.optimizer = optim.Adagrad(
				self.model.parameters(),
				lr=self.lr,
			)
		elif self.opt_method == "SGD" or self.opt_method == "sgd":
			self.optimizer = optim.SGD(
				self.model.parameters(),
				lr = self.lr,
				momentum=0.9,
			)

		if self.accelerator:
			self.optimizer = self.accelerator.prepare(self.optimizer)
			
		milestones = int(self.epochs / 3)
		self.scheduler = torch.optim.lr_scheduler.MultiStepLR(
			self.optimizer, milestones=[milestones, milestones*2],
			gamma=0.1
		)

	def train_one_step(
		self,
		data: dict[str, typing.Union[str, dgl.DGLGraph, torch.Tensor]]) -> float:

		"""根据 :py:attr:`data_loader` 生成的 1 批次（batch） ``data`` 将
		模型训练 1 步。

		:param data: 训练数据
		:type data: dict[str, typing.Union[dgl.DGLGraph, torch.Tensor]]
		:returns: 损失值
		:rtype: float
		"""

		self.optimizer.zero_grad()
		if not self.accelerator:
			data = {key : self.to_var(value) if key != 'mode' else value for key, value in data.items()}
		loss = self.model(data)
		if not self.accelerator:
			loss.backward()
		else:
			self.accelerator.backward(loss)
		self.optimizer.step()		 
		return loss.item()

	def run(self):

		"""
		训练循环，首先根据 :py:attr:`use_gpu` 设置 :py:attr:`model` 是否使用 gpu 训练，然后根据
		:py:attr:`opt_method` 设置 :py:attr:`optimizer`，最后迭代 :py:attr:`data_loader` 获取数据，
		并利用 :py:meth:`train_one_step` 训练。
		"""

		if not self.accelerator and self.use_gpu:
			self.model.cuda(device = self.device)

		if self.use_early_stopping and self.tester and self.save_path:
			self.early_stopping = EarlyStopping(
				save_path = os.path.split(self.save_path)[0],
				patience = self.patience,
				delta = self.delta)

		self.configure_optimizers()
		
		logger.info(f"[{self.get_device()}]({os.getpid()}) Initialization completed, start model training.")
		
		timer = Timer()

		for epoch in range(self.epochs):

			res = 0.0
			if not self.accelerator:
				self.model.model.train()
			else:
				self.model.module.model.train()
			
			for data in self.data_loader:
				loss = self.train_one_step(data)
				res += loss
			timer.stop()
			self.scheduler.step()
			
			if self.is_local_main_process():

				if self.valid_interval and self.tester and \
						(epoch + 1) % self.valid_interval == 0:
					logger.info(f"[{self.get_device()}]({os.getpid()}) Epoch {epoch+1} | The model starts evaluation on the validation set.")
					self.print_test("link_valid", epoch)
     
				# if self.accelerator:
				# 	self.accelerator.wait_for_everyone()
			
				if self.early_stopping and self.early_stopping.early_stop:
					logger.info(f"[{self.get_device()}]({os.getpid()}) Send an early stopping signal")
					if self.accelerator:
						self.accelerator.set_trigger()
					else:
						break

				if self.save_interval and self.save_path and (epoch + 1) % self.save_interval == 0:
					path = os.path.join(os.path.splitext(self.save_path)[0] + "-" + str(epoch+1) + \
								os.path.splitext(self.save_path)[-1])
					self.get_model().save_checkpoint(path)
					logger.info(f"[{self.get_device()}]({os.getpid()}) Epoch {epoch+1} | Training checkpoint saved at {path}")

			if self.accelerator and self.accelerator.check_trigger():
				logger.info(f"[{self.get_device()}]({os.getpid()}) Early stopping")
				break
			
			if self.log_interval and (epoch + 1) % self.log_interval == 0:
				if self.wandb_logger:
					self.wandb_logger.log({"train/train_loss" : res, "train/epoch" : epoch + 1})
				logger.info(f"[{self.get_device()}]({os.getpid()}) Epoch [{epoch+1:>4d}/{self.epochs:>4d}] | loss: {res:>9f} | {timer.avg():.5f} seconds/epoch")
		
		logger.info(f"[{self.get_device()}]({os.getpid()}) The model training is completed, taking a total of {timer.sum():.5f} seconds.")

		if self.wandb_logger:
			self.wandb_logger.log({"duration" : timer.sum()})

		if self.is_local_main_process():
      
			if self.save_path:
				self.get_model().save_checkpoint(self.save_path)
				logger.info(f"[{self.get_device()}]({os.getpid()}) Model saved at {self.save_path}.")

			if self.test and self.tester:
				logger.info(f"[{self.get_device()}]({os.getpid()}) The model starts evaluating in the test set.")
				self.print_test("link_test")

	def print_test(
		self,
		sampling_mode: str,
		epoch: int = 0):

		"""根据 :py:attr:`tester` 类型进行链接预测 。

		:param sampling_mode: 数据
		:type sampling_mode: str
		"""

		self.tester.set_sampling_mode(sampling_mode)

		if sampling_mode == "link_test":
			mode = "test"
		elif sampling_mode == "link_valid":
			mode = "val"

		results = self.tester.run_link_prediction()
		for key, value in results.items():
			logger.info(f"{key}: {value}")
		if self.wandb_logger:
			log_dict = {f"{mode}/{key}" : value for key, value in results.items()}
			if sampling_mode == "link_valid":
				log_dict.update({
					"val/epoch": epoch
				})
			self.wandb_logger.log(log_dict)
				
		if self.early_stopping is not None and sampling_mode == "link_valid":
			if self.metric in ['mr', 'mr_type']:
				self.early_stopping(-results[self.metric], self.get_model())
			elif self.metric in results.keys():
				self.early_stopping(results[self.metric], self.get_model())
			else:
				raise ValueError("Early stopping metric is not valid.")

	def to_var(
		self,
		x: torch.Tensor) -> torch.Tensor:

		"""将 ``x`` 转移到对应的设备上。

		:param x: 数据
		:type x: torch.Tensor
		:returns: 张量
		:rtype: torch.Tensor
		"""

		if self.use_gpu:
			return x.to(self.device)
		else:
			return x

	def get_model(self) -> Model:

		"""返回原始的 KGE 模型。
		
		:returns: KGE 模型
		:rtype: :py:class:`unike.module.model.Model`
		"""

		if self.accelerator:
			return self.model.module.model
		else:
			return self.model.model
		
	def get_device(self) -> typing.Union[torch.device, str]:

		"""返回当前进程的设备。
		
		:returns: 设备信息
		:rtype: typing.Union[torch.device, str]
		"""

		if self.accelerator:
			return self.model.device
		else:
			return self.device
		
	def is_local_main_process(self) -> bool:

		"""当前进程是否是主进程。
		
		:returns: 当前进程是否是主进程。
		:rtype: bool
		"""

		return not self.accelerator or self.accelerator.is_local_main_process

def get_trainer_hpo_config() -> dict[str, dict[str, typing.Any]]:

	"""返回 :py:class:`Trainer` 的默认超参数优化配置。
	
	默认配置为::
	
		parameters_dict = {
			'trainer': {
				'value': 'Trainer'
			},
			'epochs': {
				'value': 10000
			},
			'lr': {
				'distribution': 'uniform',
				'min': 1e-5,
				'max': 1.0
			},
			'opt_method': {
				'values': ['adam', 'adagrad', 'sgd']
			},
			'use_gpu_trainer': {
				'value': True
			},
			'device_trainer': {
				'value': 'cuda:0'
			},
			'valid_interval': {
				'value': 100
			},
			'log_interval': {
				'value': 100
			},
			'save_path': {
				'value': './model.pth'
			},
			'use_early_stopping': {
				'value': True
			},
			'metric': {
				'value': 'hits@10'
			},
			'patience': {
				'value': 2
			},
			'delta': {
				'value': 0.0001
			},
		}

	:returns: :py:class:`Trainer` 的默认超参数优化配置
	:rtype: dict[str, dict[str, typing.Any]]	
	"""

	parameters_dict = {
		'trainer': {
			'value': 'Trainer'
		},
		'epochs': {
			'value': 10000
		},
		'lr': {
			'distribution': 'uniform',
			'min': 1e-5,
			'max': 1.0
		},
		'opt_method': {
			'values': ['adam', 'adagrad', 'sgd']
		},
		'use_gpu_trainer': {
			'value': True
		},
		'device_trainer': {
			'value': 'cuda:0'
		},
		'valid_interval': {
			'value': 100
		},
		'log_interval': {
			'value': 100
		},
		'save_path': {
			'value': './model.pth'
		},
		'use_early_stopping': {
			'value': True
		},
		'metric': {
			'value': 'hits@10'
		},
		'patience': {
			'value': 2
		},
		'delta': {
			'value': 0.0001
		},
	}
		
	return parameters_dict


# coding:utf-8
#
# unike/data/KGEDataLoader.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 16, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Apr 27, 2024
#
# 为 KGE 模型读取数据.

"""
KGEDataLoader - KGE 模型读取数据集类。
"""

import typing
from .UniSampler import UniSampler
from .BernSampler import BernSampler
from .RGCNSampler import RGCNSampler
from .CompGCNSampler import CompGCNSampler
from .TestSampler import TestSampler
from .TradTestSampler import TradTestSampler
from torch.utils.data import DataLoader

class KGEDataLoader:

    """KGE 模型数据加载器。

    例子::

        from unike.data import KGEDataLoader, BernSampler, TradTestSampler

        dataloader = KGEDataLoader(
        	in_path = "../../benchmarks/FB15K/", 
        	batch_size = 8192,
        	neg_ent = 25,
        	test = True,
        	test_batch_size = 256,
        	num_workers = 16,
        	train_sampler = BernSampler,
        	test_sampler = TradTestSampler
        )
    """
    
    def __init__(
        self,
        in_path: str = "./",
        ent_file: str = "entity2id.txt",
        rel_file: str = "relation2id.txt",
        train_file: str = "train2id.txt",
        valid_file: str = "valid2id.txt",
        test_file: str = "test2id.txt",
        batch_size: int | None = None,
        neg_ent: int = 1,
        test: bool = False,
        test_batch_size: int | None = None,
        type_constrain: bool = True,
        num_workers: int | None = None,
        train_sampler: typing.Union[typing.Type[UniSampler], typing.Type[BernSampler], typing.Type[RGCNSampler], typing.Type[CompGCNSampler]] = BernSampler,
        test_sampler: typing.Type[TestSampler] = TradTestSampler):

        """创建 KGEDataLoader 对象。

        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param train_file: train2id.txt
        :type train_file: str
        :param valid_file: valid2id.txt
        :type valid_file: str
        :param test_file: test2id.txt
        :type test_file: str
        :param batch_size: batch size
        :type batch_size: int | None
        :param neg_ent: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity；对于 CompGCN 不起作用。
        :type neg_ent: int
        :param test: 是否读取验证集和测试集
        :type test: bool
        :param test_batch_size: test batch size
        :type test_batch_size: int | None
        :param type_constrain: 是否报告 type_constrain.txt 限制的测试结果
        :type type_constrain: bool
        :param num_workers: 加载数据的进程数
        :type num_workers: int
        :param train_sampler: 训练数据采样器
        :type train_sampler: typing.Union[typing.Type[UniSampler], typing.Type[BernSampler], typing.Type[RGCNSampler], typing.Type[CompGCNSampler]]
        :param test_sampler: 测试数据采样器
        :type test_sampler: typing.Type[TestSampler]
        """

        #: 数据集目录
        self.in_path: str = in_path
        
        if not self.in_path.endswith('/'):
            self.in_path = self.in_path + '/'
        
        #: entity2id.txt
        self.ent_file: str = ent_file
        #: relation2id.txt
        self.rel_file: str = rel_file
        #: train2id.txt
        self.train_file: str = train_file
        #: valid2id.txt
        self.valid_file: str = valid_file
        #: test2id.txt
        self.test_file: str = test_file
        #: batch size
        self.batch_size: int = batch_size
        #: 对于每一个正三元组, 构建的负三元组的个数, 替换 entity；对于 CompGCN 不起作用。
        self.neg_ent: int = neg_ent
        #: 是否读取验证集和测试集
        self.test: bool = test
        #: test batch size
        self.test_batch_size: int = test_batch_size
        #: 是否报告 type_constrain.txt 限制的测试结果
        self.type_constrain: bool = type_constrain
        #: 加载数据的进程数
        self.num_workers: int = num_workers

        #: 训练数据采样器
        self.train_sampler: typing.Union[UniSampler, BernSampler, RGCNSampler, CompGCNSampler] = train_sampler(
            in_path=self.in_path,
            ent_file=self.ent_file,
            rel_file=self.rel_file,
            train_file=self.train_file,
            batch_size=self.batch_size,
            neg_ent=self.neg_ent
        )

        #: 训练集三元组
        self.data_train: list[tuple[int, int, int]] = self.train_sampler.get_train()
        
        self.validate_data_train()

        if self.test:
            #: 测试数据采样器
            self.test_sampler: TestSampler = test_sampler(
                sampler=self.train_sampler,
                valid_file=self.valid_file,
                test_file=self.test_file,
                type_constrain=type_constrain
            )
        
            #: 验证集三元组
            self.data_val: list[tuple[int, int, int]] = self.test_sampler.get_valid()
            #: 测试集三元组
            self.data_test: list[tuple[int, int, int]] = self.test_sampler.get_test()
            
            self.validate_data_test()
            
    def validate_data_train(self) -> None:
        """验证训练集中的实体和关系 id 是否合法。"""
        
        ent_tol = self.get_ent_tol()
        rel_tol = self.get_rel_tol()
        
        for h, r, t in self.data_train:
            if h >= ent_tol or r >= rel_tol or t >= ent_tol:
                raise ValueError(f"the entity id or relation id of train data is out of range: {h}, {r}, {t}")
    
    def validate_data_test(self) -> None:
        """验证测试集和验证集中的实体和关系 id 是否合法。"""
        
        ent_tol = self.get_ent_tol()
        rel_tol = self.get_rel_tol()
        
        for h, r, t in self.data_val:
            if h >= ent_tol or r >= rel_tol or t >= ent_tol:
                raise ValueError(f"the entity id or relation id of test data is out of range: {h}, {r}, {t}")
        for h, r, t in self.data_test:
            if h >= ent_tol or r >= rel_tol or t >= ent_tol:
                raise ValueError(f"the entity id or relation id of test data is out of range: {h}, {r}, {t}")
        
    def get_ent_tol(self) -> int:

        """返回实体个数。
        
        :returns: 实体个数
        :rtype: int
        """

        return self.train_sampler.ent_tol

    def get_rel_tol(self) -> int:

        """返回关系个数。
        
        :returns: 关系个数
        :rtype: int
        """

        return self.train_sampler.rel_tol

    def train_dataloader(self) -> DataLoader:

        """返回训练数据加载器。
        
        :returns: 训练数据加载器
        :rtype: torch.utils.data.DataLoader
        """

        return DataLoader(
            self.data_train,
            shuffle=True,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
            drop_last=True,
            collate_fn=self.train_sampler.sampling,
        )
            
    def val_dataloader(self) -> DataLoader:

        """返回验证数据加载器。
        
        :returns: 验证数据加载器
        :rtype: torch.utils.data.DataLoader
        """

        return DataLoader(
            self.data_val,
            shuffle=False,
            batch_size=self.test_batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
            collate_fn=self.test_sampler.sampling,
        )

    def test_dataloader(self) -> DataLoader:

        """返回测试数据加载器。
        
        :returns: 测试数据加载器
        :rtype: torch.utils.data.DataLoader"""
        
        return DataLoader(
            self.data_test,
            shuffle=False,
            batch_size=self.test_batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
            collate_fn=self.test_sampler.sampling,
        )

def get_kge_data_loader_hpo_config() -> dict[str, dict[str, typing.Any]]:
    
    """返回 :py:class:`KGEDataLoader` 的默认超参数优化配置。
    
    默认配置为::
    
        parameters_dict = {
            'dataloader': {
                'value': 'KGEDataLoader'
            },
            'in_path': {
                'value': './'
            },
            'ent_file': {
                'value': 'entity2id.txt'
            },
            'rel_file': {
                'value': 'relation2id.txt'
            },
            'train_file': {
                'value': 'train2id.txt'
            },
            'valid_file': {
                'value': 'valid2id.txt'
            },
            'test_file': {
                'value': 'test2id.txt'
            },
            'batch_size': {
                'values': [512, 1024, 2048, 4096]
            },
            'neg_ent': {
                'values': [1, 4, 16, 64]
            },
            'test_batch_size': {
                'value': 30
            },
            'type_constrain': {
                'value': True
            },
            'num_workers': {
                'value': 16
            },
            'train_sampler': {
                'value': 'BernSampler'
            },
            'test_sampler': {
                'value': 'TradTestSampler'
            }
        }
    
    :returns: :py:class:`KGEDataLoader` 的默认超参数优化配置
    :rtype: dict[str, dict[str, typing.Any]]
    """
    
    parameters_dict = {
        'dataloader': {
            'value': 'KGEDataLoader'
        },
        'in_path': {
            'value': './'
        },
        'ent_file': {
            'value': 'entity2id.txt'
        },
        'rel_file': {
            'value': 'relation2id.txt'
        },
        'train_file': {
            'value': 'train2id.txt'
        },
        'valid_file': {
            'value': 'valid2id.txt'
        },
        'test_file': {
            'value': 'test2id.txt'
        },
        'batch_size': {
            'values': [512, 1024, 2048, 4096]
        },
        'neg_ent': {
            'values': [1, 4, 16, 64]
        },
        'test_batch_size': {
            'value': 30
        },
        'type_constrain': {
            'value': True
        },
        'num_workers': {
            'value': 16
        },
        'train_sampler': {
            'value': 'BernSampler'
        },
        'test_sampler': {
            'value': 'TradTestSampler'
        }
    }
    
    return parameters_dict


# coding:utf-8
#
# unike/data/__init__.py
# 
# git pull from OpenKE-PyTorch by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on May 7, 2023
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 29, 2023
# 
# 该头文件定义了 data 接口.

"""数据采样部分，包含为训练和验证模型定义的数据采样器。"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .KGReader import KGReader

from .TradSampler import TradSampler
from .UniSampler import UniSampler
from .BernSampler import BernSampler

from .RevSampler import RevSampler
from .RGCNSampler import RGCNSampler
from .CompGCNSampler import CompGCNSampler

from .TestSampler import TestSampler
from .TradTestSampler import TradTestSampler
from .RGCNTestSampler import RGCNTestSampler
from .CompGCNTestSampler import CompGCNTestSampler

from .KGEDataLoader import KGEDataLoader, get_kge_data_loader_hpo_config

__all__ = [
	'KGReader',
	'TradSampler',
	'UniSampler',
	'BernSampler',
	'RevSampler',
	'RGCNSampler',
	'CompGCNSampler',
	'TestSampler',
	'TradTestSampler',
	'RGCNTestSampler',
	'CompGCNTestSampler',
	'KGEDataLoader',
	'get_kge_data_loader_hpo_config'
]


# coding:utf-8
#
# unike/utils/Link.py
#
# created by wangtao <wangtao.cpu@gmail.com> on Jul 10, 2025
#
# 该脚本定义了 Link 类.

"""
Link - 链接分数计算。
"""

import torch
import pandas as pd
import os
from ..module.model import Model

class Link:
    """
    使用 KGE 模型对感兴趣的三元组计算链接分数。
    """
    def __init__(self,
                in_path: str = './',
                ent_file: str = "entity2id.txt", 
                rel_file: str = "relation2id.txt",
                all_file: str = "all2id.txt"):
        
        """创建 Link 对象。
        
        :param model: 模型
        :type model: Model
        :param in_path: 数据集目录
        :type in_path: str
        :param ent_file: entity2id.txt
        :type ent_file: str
        :param rel_file: relation2id.txt
        :type rel_file: str
        :param all_file: all2id.txt
        :type all_file: str
        """

        self.in_path = in_path
        self.ent_file = ent_file
        self.rel_file = rel_file
        self.all_file = all_file
        
        self.load_data()
        self.load_all()
        
        
    def load_data(self) -> None:
        
        """读取 :py:attr:`ent_file` 文件和 :py:attr:`rel_file` 文件。"""
        
        self.ent2id = {}
        self.id2ent = {}

        with open(os.path.join(self.in_path, self.ent_file)) as f:
            _ = (int)(f.readline())
            for line in f:
                entity, id_ = line.strip().split("\t")
                self.ent2id[entity] = int(id_)
                self.id2ent[int(id_)] = entity

        self.rel2id = {}
        self.id2rel = {}

        with open(os.path.join(self.in_path, self.rel_file)) as f:
            _ = (int)(f.readline())
            for line in f:
                relation, id_ = line.strip().split("\t")
                self.rel2id[relation] = int(id_)
                self.id2rel[int(id_)] = relation
    
    def load_all(self) -> None:
        
        """读取 :py:attr:`all_file` 文件。"""
        
        self.all = set()
        with open(os.path.join(self.in_path, self.all_file)) as f:
            _ = f.readline()
            for line in f:
                head, tail, rel = line.strip().split()
                self.all.add((int(head), int(rel), int(tail)))
        
        
    def link(self, 
             head_ids: list[int], 
             rel_ids: list[int], 
             tail_ids: list[int], 
             model: Model, 
             device: str = 'cpu'
        ) -> pd.DataFrame:
        
        """对给定的头实体、关系和尾实体进行组合并使用模型计算链接分数。
        
        :param head_ids: 头实体列表
        :type head_ids: list[int]
        :param rel_ids: 关系列表
        :type rel_ids: list[int]
        :param tail_ids: 尾实体列表
        :type tail_ids: list[int]
        :param device: 模型
        :type device: Model
        :param device: 设备
        :type device: str
        :returns: 结果数据框
        :rtype: pd.DataFrame
        """
        
        head_ids_ = torch.tensor(head_ids).long().to(device)
        rel_ids_ = torch.tensor(rel_ids).long().to(device)
        tail_ids_ = torch.tensor(tail_ids).long().to(device)

        triples = []
        scores = []
        
        model.eval()
        model.to(device)
        with torch.no_grad():
            for h_idx in range(len(head_ids_)):
                for r_idx in range(len(rel_ids_)):
                    h_id = head_ids_[h_idx]
                    r_id = rel_ids_[r_idx]
                    
                    h_id = h_id.tile((tail_ids_.shape[0], ))
                    r_id = r_id.tile((tail_ids_.shape[0], ))
                    
                    triple = torch.stack((h_id, r_id, tail_ids_)).T
                    data = {
                        "positive_sample": triple
                    }
                    score = model.predict(data, "single")
                    
                    scores.append(score)
                    triples.append(triple)
        
        triples = torch.cat(triples)
        scores = torch.cat(scores)
        
        sorted_scores, indices = torch.sort(scores.squeeze(), descending=True)
        sorted_triples = triples[indices]

        df = pd.DataFrame(sorted_triples.tolist(), columns=["head", "rel", "tail"])
        df["score"] = sorted_scores.tolist()
        df["in"] = df.apply(lambda row: (row['head'], row['rel'], row['tail']) in self.all, axis=1)
        df["head_ent"] = df["head"].map(self.id2ent)
        df["rel_ent"] = df["rel"].map(self.id2rel)
        df["tail_ent"] = df["tail"].map(self.id2ent)
        
        return df



# coding:utf-8
#
# unike/utils/WandbLogger.py
#
# created by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Jan 1, 2024
# updated by LuYF-Lemon-love <luyanfeng_nlp@qq.com> on Feb 24, 2024
#
# 该脚本定义了 WandbLogger 类.

"""
WandbLogger - 使用 Weights and Biases 记录实验结果。
"""

import wandb
import swanlab
from typing import Literal, Any, Optional
from accelerate import Accelerator
from types import SimpleNamespace


class WandbLogger:
    """使用 `Weights and Biases <https://docs.wandb.ai/>`_ 记录实验结果。"""

    def __init__(self, endpoint: Literal['wandb', 'swanlab'] = 'wandb'):

        """创建 WandbLogger 对象。
        
        :param endpoint: 使用 wandb 还是 swanlab 记录实验结果
        :type endpoint: Literal['wandb', 'swanlab']
        """
        self.config: SimpleNamespace = SimpleNamespace()
        self.endpoint = endpoint
        self.logger: Any = None

        self.project: str = 'UniKE'
        self.name: str = ''
        self.offline = False

    def set_config(self, project: str, name: str, config: dict[str, Any] | None = None, offline: bool = False) -> 'WandbLogger':
        """设置项目名称和配置。
        
        :param project: wandb 的项目名称
        :type project: str
        :param name: wandb 的 run name
        :type name: str
        :param config: wandb 的项目配置如超参数。
        :type config: dict[str, Any] | None
        :param offline: 是否离线模式
        :type offline: bool
        """
        self.project = project
        self.name = name
        self.offline = offline
        if config:
            self.config: SimpleNamespace = SimpleNamespace(**config)
        return self

    def _init(self):
        if self.logger is not None:
            return
        if self.endpoint == 'wandb':
            if not self.offline:
                wandb.login()
            wandb.init(project=self.project, name=self.name, config=self.config.__dict__, mode='offline' if self.offline else 'online')
            self.logger = wandb
        elif self.endpoint == 'swanlab':
            if not self.offline:
                swanlab.login()
            swanlab.init(project=self.project, name=self.name, config=self.config.__dict__, mode='offline' if self.offline else 'online')
            self.logger = swanlab

    def log(self, *args, **kwargs):
        """记录日志"""
        self.logger.log(*args, **kwargs)

    def finish(self):
        """关闭日志"""
        if isinstance(self.logger, Accelerator):
            self.logger.end_training()
        else:
            self.logger.finish()
